import os
from typing import Callable, Optional, Tuple, Any, Dict
from fuse.data.ops.ops_cast import Cast
from fuse.data.utils.sample import get_sample_id
from fuse.utils import NDict
from fuse.utils.file_io.file_io import create_dir
from fuse.data.ops.ops_debug import OpDebugBase
import numpy
import torch
import matplotlib.pyplot as plt
import SimpleITK as sitk
import plotly.graph_objects as go
import numpy as np


class OpVis2DImage(OpDebugBase):
    """
    Visualize a 2D image (either display it or save it as an image).
    Supports both rgb and gray scale and supports few image formats "channels_first"/"no_channels"/"channels_last"
    It's recommended, but not a must, to run it in a single process.
    Add at the top your script to force single process:
    ```
    from fuse.utils.utils_debug import FuseDebug
    FuseDebug("debug")
    ```
    Example:
    # Display RGB image [3, H, W] already in range [0-256]
    (OpVis2DImage(num_samples=1), dict(key="data.input.img", dtype="int"))

    """

    def __init__(
        self,
        show: bool = True,
        path: str = ".",
        image_format: str = "channels_first",
        image_process_func: Optional[Callable] = None,
        **kwargs: dict,
    ):
        """
        :param show: if set to true will display it, otherwise will save it in `path`
        :param path: location to save the images. Used when `show` set to False
        :param image_format: the format of the image: "channels_first"/"no_channels"/"channels_last"
        :param image_processing_func: callable that get the image (numpy array) and can process it. Typically use to move the range to values matplotlib can display.
        :param kwargs: see super class arguments
        """
        super().__init__(**kwargs)
        self._path = path
        self._image_format = image_format
        self._image_process_func = image_process_func
        self._show = show

    def call_debug(
        self,
        sample_dict: NDict,
        key: str,
        clip: Optional[Tuple] = None,
        dtype: Optional[str] = None,
        figure_kwargs: dict = {},
        **imshow_kwargs: Dict[str, Any],
    ) -> None:
        """
        :param key: sample_dict key to a 2D image. Either tensor or numpy array.
        :param clip: tuple of arguments to numpy.clip if the operation required
        :param dtype: in case the range is [0-256], matplotlib expects int. Set to "int" if this is the case.
        :param figure_kwargs: parameters to pass to plt.figure
        :params imshow_kwargs: extra parameters to pass to plt.imshow
        """

        img = sample_dict[key]
        if isinstance(img, torch.Tensor):
            img = img.detach().cpu().numpy()

        assert isinstance(img, numpy.ndarray)

        if clip is not None:
            img = img.clip(*clip)

        if dtype is not None:
            img = Cast.to_numpy(img, dtype=dtype)

        if self._image_format == "channels_first":
            img = numpy.moveaxis(img, 0, -1)
        elif self._image_format == "no_channels":
            img = numpy.expand_dims(img, axis=-1)

        if self._image_process_func is not None:
            img = self._image_process_func(img)

        if "cmap" not in imshow_kwargs:
            if img.shape[-1] == 1:  # assume and force gray scale
                imshow_kwargs["cmap"] = "gray"

        plt.figure(**figure_kwargs)
        plt.imshow(img, **imshow_kwargs)
        if self._show:
            plt.show()
        else:
            if self._name is None:
                filename = (
                    os.path.join(
                        self._path, get_sample_id(sample_dict).replace(".", "__")
                    )
                    + ".png"
                )
            else:
                filename = (
                    os.path.join(
                        self._path,
                        self._name,
                        get_sample_id(sample_dict).replace(".", "__"),
                    )
                    + ".png"
                )
            create_dir(os.path.dirname(filename))
            plt.savefig(filename)  # save fig


class OpVisImageHist(OpDebugBase):
    """
    Visualize single image histogram (either display it or save it as an image).
    ```
    from fuse.utils.utils_debug import FuseDebug
    FuseDebug("debug")
    ```
    Example:
    ```
    (OpVisImageHist(first_sample_only=True), dict(key="data.input.img"))
    ```
    """

    def __init__(self, show: bool = True, path: str = ".", **kwargs: dict):
        """
        :param show: if set to true will display it, otherwise will save it in `path`
        :param path: location to save the images. Used when `show` set to False
        :param kwargs: see super class arguments
        """
        super().__init__(**kwargs)
        self._path = path
        self._show = show

    def call_debug(
        self,
        sample_dict: NDict,
        key: str,
        bins: int = 10,
        figure_kwargs: dict = {},
        **hist_kwargs: dict,
    ) -> NDict:
        """
        :param key: sample_dict key to a 2D image. Either tensor or numpy array.
        :param figure_kwargs: parameters to pass to plt.figure
        :params hist_kwargs: extra parameters to pass to plt.hist
        """

        img = sample_dict[key]
        if isinstance(img, torch.Tensor):
            img = img.detach().cpu().numpy()

        assert isinstance(img, numpy.ndarray)

        if self._image_format == "channels_first":
            img = numpy.moveaxis(img, 0, -1)
        if self._image_process_func is not None:
            img = self._image_process_func(img)

        plt.figure(**figure_kwargs)
        plt.hist(img.flatten(), bins=bins, **hist_kwargs)
        if self._show:
            plt.show()
        else:
            plt.savefig(
                os.path.join(
                    self._path,
                    f"hist_{get_sample_id(sample_dict).replace('.', '__')}.png",
                )
            )  # save fig


class OpVis3DPlotly(OpDebugBase):
    """
    Visualize a 3D image as plotly figure exported to html.
    Supports only gray scale.
    To view results, open the <path>_PLOTLY.html file generated in chrome (have not yet tested another browser).
    Example of use:
            (OpVis3DPlotly(num_samples=1), dict(key="data.debug.3d_volume"))
    """

    def __init__(
        self,
        path: str = ".",
        callback: Optional[Callable] = None,
        **kwargs: Dict[str, Any],
    ):
        """
        :param path: location to save .html rendered volume files
        :param callback: function which accepts np.ndarray which can be applied immediately before visualization to the volume
        :param kwargs: see super class arguments
        Example using callback (with some function to binarize the volume):
            (OpVis3DPlotly(num_samples=1, callback=lambda x:np.where(x>0.5, 1, 0)), dict(key="data.debug.3d_volume")),
        """
        super().__init__(**kwargs)
        self._path = path
        self._callback = callback

    def frame_args(self, duration: int) -> dict:
        """
        Required for buttons used in the plotly graph animation.
        Everything is hardcoded/constant except for duration.
        :param duration: length of animation once relevant button is activated
        """

        return {
            "frame": {"duration": duration},
            "mode": "immediate",
            "fromcurrent": True,
            "transition": {"duration": duration, "easing": "linear"},
        }

    def get_plotly_fig_from_vol(self, vol: np.ndarray) -> go.Figure:
        """
        this function will return a plotly figure (which can later be saved) created from a 3d numpy-like matrix
        :param vol: lxwxh numpy-like matrix
        returns: plotly.graph_objs._figure.Figure
        """
        # uncomment the below line for example from their website (brain MRI)
        # vol = io.imread("<https://s3.amazonaws.com/assets.datacamp.com/blog_assets/attention-mri.tif>")

        nb_frames, w, h = vol.shape
        cmin = vol.min()
        cmax = vol.max()

        fig = go.Figure(
            frames=[
                go.Frame(
                    data=go.Surface(
                        z=(nb_frames - k) * np.ones((w, h)),  # height
                        surfacecolor=np.flipud(vol[(nb_frames - 1) - k]),
                        cmin=cmin,
                        cmax=cmax,
                    ),
                    name=str(
                        k
                    ),  # you need to name the frame for the animation to behave properly
                )
                for k in range(nb_frames)
            ]
        )

        # Add data to be displayed before animation starts
        fig.add_trace(
            go.Surface(
                z=nb_frames * np.ones((w, h)),
                surfacecolor=np.flipud(vol[0]),
                colorscale="Gray",
                cmin=cmin,
                cmax=cmax,
                colorbar=dict(thickness=20, ticklen=4),
            )
        )

        sliders = [
            {
                "pad": {"b": 10, "t": 60},
                "len": 0.9,
                "x": 0.1,
                "y": 0,
                "steps": [
                    {
                        "args": [[f.name], self.frame_args(0)],
                        "label": str(k),
                        "method": "animate",
                    }
                    for k, f in enumerate(fig.frames)
                ],
            }
        ]

        # Layout
        fig.update_layout(
            title="Slices in volumetric data",
            width=600,
            height=600,
            scene=dict(
                zaxis=dict(range=[-1, nb_frames], autorange=False),
                aspectratio=dict(x=1, y=1, z=1),
            ),
            updatemenus=[
                {
                    "buttons": [
                        {
                            "args": [None, self.frame_args(50)],
                            "label": "&#9654;",  # play symbol
                            "method": "animate",
                        },
                        {
                            "args": [[None], self.frame_args(0)],
                            "label": "&#9724;",  # pause symbol
                            "method": "animate",
                        },
                    ],
                    "direction": "left",
                    "pad": {"r": 10, "t": 70},
                    "type": "buttons",
                    "x": 0.1,
                    "y": 0,
                }
            ],
            sliders=sliders,
        )
        return fig

    def save_plotly_fig_html(self, fig: go.Figure, sample_dict: NDict) -> None:
        """
        Saves a plotly figure object as html.
        :param fig: the plotly figure object that you want to save
        :param sample_dict: dict containing minibatch sample. Used only to extract sample id when saving to filename
        """
        if self._name is None:
            # filename will be something like ./ProstateX-0008_1_PLOTLY.html
            filename = (
                os.path.join(self._path, get_sample_id(sample_dict).replace(".", "__"))
                + "_PLOTLY"
                + ".html"
            )
        else:
            filename = (
                os.path.join(
                    self._path,
                    self._name,
                    get_sample_id(sample_dict).replace(".", "__"),
                )
                + "_PLOTLY"
                + ".html"
            )
        create_dir(os.path.dirname(filename))
        fig.write_html(filename)

    def call_debug(
        self,
        sample_dict: NDict,
        key: str,
    ) -> None:
        """
        :param sample_dict: dict containing minibatch sample
        :param key: sample_dict key to a 3D image. Either tensor or numpy array.
        """

        # ensure vol is np.ndarray
        vol = sample_dict[key]
        if isinstance(vol, torch.Tensor):
            vol = vol.detach().cpu().numpy()
        if isinstance(vol, sitk.Image):  # should not be array! temp sagi
            vol = sitk.GetArrayFromImage(vol)
        assert isinstance(vol, numpy.ndarray)

        if self._callback is not None:
            vol = self._callback(vol)

        # make plotly figure and save
        plotly_fig = self.get_plotly_fig_from_vol(vol)
        self.save_plotly_fig_html(plotly_fig, sample_dict)



import unittest

from typing import Optional, OrderedDict, Union, List, Any

import torch

from fuse.utils.ndict import NDict

from fuse.data.ops.op_base import OpBase, OpReversibleBase, op_call
from fuse.data.key_types_for_testing import DataTypeForTesting

from fuse.data.ops.ops_common import (
    OpApplyPatterns,
    OpFunc,
    OpLambda,
    OpRepeat,
    OpReplaceAnyElements,
    OpReplaceElements,
)
from fuse.data.ops.ops_common_for_testing import OpApplyTypesImaging


class OpIncrForTest(OpReversibleBase):
    def __init__(self, **kwargs: dict):
        super().__init__(**kwargs)

    def __call__(
        self,
        sample_dict: NDict,
        op_id: Optional[str],
        incr_value: int,
        key_in: str,
        key_out: str,
    ) -> Union[None, dict, List[dict]]:
        # save for reverse
        sample_dict[op_id] = {"key_out": key_out, "incr_value": incr_value}
        # apply
        value = sample_dict[key_in]
        sample_dict[key_out] = value + incr_value

        return sample_dict

    def reverse(
        self,
        sample_dict: NDict,
        key_to_reverse: str,
        key_to_follow: str,
        op_id: Optional[str],
    ) -> dict:
        # not really reverse, but help the test
        orig_args = sample_dict[op_id]

        if orig_args["key_out"] != key_to_follow:
            return sample_dict

        value = sample_dict[key_to_reverse]
        sample_dict[key_to_reverse] = value - orig_args["incr_value"]

        return sample_dict


class TestOpsCommon(unittest.TestCase):
    def test_op_repeat(self) -> None:
        """
        Test OpRepeat __call__() and reverse()
        """
        op_base = OpIncrForTest()
        kwargs_per_step_to_add = [
            dict(key_in="data.val.a", key_out="data.val.b"),
            dict(key_in="data.val.b", key_out="data.val.c"),
            dict(key_in="data.val.b", key_out="data.val.d"),
            dict(key_in="data.val.d", key_out="data.val.d"),
        ]
        op_repeat = OpRepeat(op_base, kwargs_per_step_to_add)
        sample_dict = NDict({})
        sample_dict["data.val.a"] = 5
        sample_dict = op_repeat(sample_dict, "_.test_repeat", incr_value=3)
        self.assertEqual(sample_dict["data.val.a"], 5)
        self.assertEqual(sample_dict["data.val.b"], 8)
        self.assertEqual(sample_dict["data.val.c"], 11)
        self.assertEqual(sample_dict["data.val.d"], 14)

        op_repeat.reverse(
            sample_dict,
            key_to_follow="data.val.d",
            key_to_reverse="data.val.d",
            op_id="_.test_repeat",
        )
        self.assertEqual(sample_dict["data.val.a"], 5)
        self.assertEqual(sample_dict["data.val.b"], 8)
        self.assertEqual(sample_dict["data.val.c"], 11)
        self.assertEqual(sample_dict["data.val.d"], 8)

        sample_dict["data.val.e"] = 48
        op_repeat.reverse(
            sample_dict,
            key_to_follow="data.val.d",
            key_to_reverse="data.val.e",
            op_id="_.test_repeat",
        )
        self.assertEqual(sample_dict["data.val.a"], 5)
        self.assertEqual(sample_dict["data.val.b"], 8)
        self.assertEqual(sample_dict["data.val.c"], 11)
        self.assertEqual(sample_dict["data.val.d"], 8)
        self.assertEqual(sample_dict["data.val.e"], 42)

    def test_op_lambda(self) -> None:
        """
        Test OpLambda __call__() and reverse()
        """
        op_base = OpLambda(func=lambda x: x + 3)
        kwargs_per_step_to_add = [dict(), dict(), dict()]
        op_repeat = OpRepeat(op_base, kwargs_per_step_to_add)
        sample_dict = NDict({})
        sample_dict["data.val.a"] = 5
        sample_dict = op_repeat(sample_dict, "_.test_repeat", key="data.val.a")
        self.assertEqual(sample_dict["data.val.a"], 14)

        op_base = OpLambda(func=lambda x: x + 3, func_reverse=lambda x: x - 3)
        op_repeat = OpRepeat(op_base, kwargs_per_step_to_add)
        sample_dict = NDict({})
        sample_dict["data.val.a"] = 5
        sample_dict = op_repeat(sample_dict, "_.test_repeat", key="data.val.a")
        self.assertEqual(sample_dict["data.val.a"], 14)

        op_repeat.reverse(
            sample_dict,
            key_to_follow="data.val.a",
            key_to_reverse="data.val.a",
            op_id="_.test_repeat",
        )
        self.assertEqual(sample_dict["data.val.a"], 5)

        sample_dict["data.val.b"] = 51
        op_repeat.reverse(
            sample_dict,
            key_to_follow="data.val.a",
            key_to_reverse="data.val.b",
            op_id="_.test_repeat",
        )
        self.assertEqual(sample_dict["data.val.a"], 5)
        self.assertEqual(sample_dict["data.val.b"], 42)

    def test_op_lambda_with_kwargs(self) -> None:
        """
        Test OpLambda __call__() with kwargs
        """
        op_base = OpLambda(func=lambda x, y: x + y)
        kwargs_per_step_to_add = [dict(), dict(), dict()]
        op_repeat = OpRepeat(op_base, kwargs_per_step_to_add)
        sample_dict = NDict()
        sample_dict["data.val.a"] = 5
        sample_dict = op_repeat(sample_dict, "_.test_repeat", key="data.val.a", y=5)
        self.assertEqual(sample_dict["data.val.a"], 20)

    def test_op_func(self) -> None:
        """
        Test OpFunc __call__()
        """

        def func_single_output(a, b, c):  # type: ignore
            return a + b + c

        def func_multi_output(a, b, c):  # type: ignore
            return a + b, a + c

        single_output_op = OpFunc(func=func_single_output)
        sample_dict = NDict({})
        sample_dict["data.first"] = 5
        sample_dict["data.second"] = 9
        sample_dict = op_call(
            single_output_op,
            sample_dict,
            "_.test_func",
            c=2,
            inputs={"data.first": "a", "data.second": "b"},
            outputs="data.out",
        )
        self.assertEqual(sample_dict["data.out"], 16)

        multi_output_op = OpFunc(func=func_multi_output)
        sample_dict = NDict({})
        sample_dict["data.first"] = 5
        sample_dict["data.second"] = 9
        sample_dict = op_call(
            multi_output_op,
            sample_dict,
            "_.test_func",
            c=2,
            inputs={"data.first": "a", "data.second": "b"},
            outputs=["data.out", "data.more"],
        )
        self.assertEqual(sample_dict["data.out"], 14)
        self.assertEqual(sample_dict["data.more"], 7)

    def test_op_apply_patterns(self) -> None:
        """
        Test OpRApplyPatterns __call__() and reverse()
        """

        op_add_1 = OpLambda(func=lambda x: x + 1, func_reverse=lambda x: x - 1)
        op_mul_2 = OpLambda(func=lambda x: x * 2, func_reverse=lambda x: x // 2)
        op_mul_4 = OpLambda(func=lambda x: x * 4, func_reverse=lambda x: x // 4)

        sample_dict = NDict({})
        sample_dict["data.val.img_for_testing"] = 3
        sample_dict["data.test.img_for_testing"] = 3
        sample_dict["data.test.seg_for_testing"] = 3
        sample_dict["data.test.bbox_for_testing"] = 3
        sample_dict["data.test.meta"] = 3

        patterns_dict = OrderedDict(
            [
                (r"^data.val.img_for_testing$", (op_add_1, dict())),
                (r"^.*img_for_testing$|^.*seg_for_testing$", (op_mul_2, dict())),
                (r"^data.[^.]*.bbox_for_testing", (op_mul_4, dict())),
            ]
        )
        op_apply_pat = OpApplyPatterns(patterns_dict)

        sample_dict = op_apply_pat(sample_dict, "_.test_apply_pat")
        self.assertEqual(sample_dict["data.val.img_for_testing"], 4)
        self.assertEqual(sample_dict["data.test.img_for_testing"], 6)
        self.assertEqual(sample_dict["data.test.seg_for_testing"], 6)
        self.assertEqual(sample_dict["data.test.bbox_for_testing"], 12)
        self.assertEqual(sample_dict["data.test.meta"], 3)

        sample_dict["model.seg_for_testing"] = 3
        op_apply_pat.reverse(
            sample_dict,
            key_to_follow="data.val.img_for_testing",
            key_to_reverse="model.seg_for_testing",
            op_id="_.test_apply_pat",
        )
        self.assertEqual(sample_dict["data.val.img_for_testing"], 4)
        self.assertEqual(sample_dict["model.seg_for_testing"], 2)

    def test_op_apply_types(self) -> None:
        """
        Test OpApplyTypes __call__() and reverse()
        """

        op_add_1 = OpLambda(func=lambda x: x + 1, func_reverse=lambda x: x - 1)
        op_mul_2 = OpLambda(func=lambda x: x * 2, func_reverse=lambda x: x // 2)
        op_mul_4 = OpLambda(func=lambda x: x * 4, func_reverse=lambda x: x // 4)

        sample_dict = NDict({})
        sample_dict["data.val.img_for_testing"] = 3
        sample_dict["data.test.img_for_testing"] = 3
        sample_dict["data.test.seg_for_testing"] = 3
        sample_dict["data.test.bbox_for_testing"] = 3
        sample_dict["data.test.meta"] = 3

        types_dict = {
            DataTypeForTesting.IMAGE_FOR_TESTING: (op_add_1, dict()),
            DataTypeForTesting.SEG_FOR_TESTING: (op_mul_2, dict()),
            DataTypeForTesting.BBOX_FOR_TESTING: (op_mul_4, dict()),
        }

        op_apply_type = OpApplyTypesImaging(types_dict)

        sample_dict = op_apply_type(sample_dict, "_.test_apply_type")
        self.assertEqual(sample_dict["data.val.img_for_testing"], 4)
        self.assertEqual(sample_dict["data.test.img_for_testing"], 4)
        self.assertEqual(sample_dict["data.test.seg_for_testing"], 6)
        self.assertEqual(sample_dict["data.test.bbox_for_testing"], 12)
        self.assertEqual(sample_dict["data.test.meta"], 3)

        sample_dict["model.a_seg_for_testing"] = 3
        op_apply_type.reverse(
            sample_dict,
            key_to_follow="data.val.img_for_testing",
            key_to_reverse="model.a_seg_for_testing",
            op_id="_.test_apply_type",
        )
        self.assertEqual(sample_dict["data.val.img_for_testing"], 4)
        self.assertEqual(sample_dict["model.a_seg_for_testing"], 2)

    def test_op_replace_element(self) -> None:
        """
        test op_replace_element on each of possible inputs it supports: string, list, tensor
        """

        def _apply_op_replace(input: Any, to_replace: Any, replace_with: str) -> Any:
            sample_dict = NDict({})
            sample_dict["data.input"] = input
            sample_dict = op_call(
                OpReplaceElements(),
                sample_dict,
                "_.test_replace_element",
                key_in="data.input",
                key_out="data.out",
                find_val=to_replace,
                replace_with_val=replace_with,
            )
            return sample_dict["data.out"]

        # input is a string
        input = "abccba"
        to_replace = "b"
        replace_with = "X"
        expected_result = "aXccXa"

        res = _apply_op_replace(input, to_replace, replace_with)
        self.assertEqual(res, expected_result)

        # input is a list
        input = [1, 2, 3, 2, 1]
        to_replace = 2
        replace_with = 0
        expected_result = [1, 0, 3, 0, 1]
        res = _apply_op_replace(input, to_replace, replace_with)
        self.assertEqual(res, expected_result)

        # input is a tensor
        res = _apply_op_replace(torch.tensor(input), to_replace, replace_with)
        self.assertTrue(torch.equal(res, torch.tensor(expected_result)))

    def test_op_replace_any_element(self) -> None:
        """
        test op_replace_any_element on each of possible inputs it supports: string, list, tensor
        """

        def _apply_op_replace_any(
            input: Any, to_replace: Any, replace_with: str
        ) -> Any:
            sample_dict = NDict({})
            sample_dict["data.input"] = input
            sample_dict = op_call(
                OpReplaceAnyElements(),
                sample_dict,
                "_.test_replace_element",
                key_in="data.input",
                key_out="data.out",
                find_any_val=to_replace,
                replace_with_val=replace_with,
            )
            return sample_dict["data.out"]

        # input is a string
        input = "abccba"
        to_replace = "bc"
        replace_with = "X"
        expected_result = "aXXXXa"

        res = _apply_op_replace_any(input, to_replace, replace_with)
        self.assertEqual(res, expected_result)

        # input is a list
        input = [1, 2, 3, 2, 1]
        to_replace = [2, 3]
        replace_with = 0
        expected_result = [1, 0, 0, 0, 1]
        res = _apply_op_replace_any(input, to_replace, replace_with)
        self.assertEqual(res, expected_result)

        # input is a tensor
        res = _apply_op_replace_any(torch.tensor(input), to_replace, replace_with)
        self.assertTrue(torch.equal(res, torch.tensor(expected_result)))


if __name__ == "__main__":
    unittest.main()



from typing import Union, Optional
from fuseimg.datasets.mnist import MNIST
from typing import Sequence
from fuse.data import DatasetDefault


def create_dataset(
    train_val_sample_ids: Union[Sequence, None] = None,
    paths: Optional[dict] = None,
    params: Optional[dict] = None,
) -> Sequence[DatasetDefault]:
    train_val_dataset = MNIST.dataset(paths["cache_dir"], train=True)
    if train_val_sample_ids is None:
        test_dataset = MNIST.dataset(paths["cache_dir"], train=False)
        return train_val_dataset, test_dataset
    else:
        train_dataset = MNIST.dataset(
            paths["cache_dir"], train=True, sample_ids=train_val_sample_ids[0]
        )
        validation_dataset = MNIST.dataset(
            paths["cache_dir"], train=True, sample_ids=train_val_sample_ids[1]
        )
        return train_dataset, validation_dataset



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021
"""

import pytorch_lightning as pl
from typing import Optional, Union, Tuple, Callable
from collections import OrderedDict
import os

from fuse.dl.lightning.pl_funcs import *  # noqa
from fuse.utils.file_io.file_io import create_dir


class LightningModuleDefault(pl.LightningModule):
    """
    Generic implementation of LightningModule using FuseMedML style focusing primarily on supervised training.
    FuseMedML conventions make it possible to have such a generic implementation.
    """

    def __init__(
        self,
        model_dir: Optional[str],
        model: torch.nn.Module,
        losses: Optional[Dict[str, LossBase]] = None,
        validation_losses: Optional[List[Tuple[str, Dict[str, LossBase]]]] = None,
        train_metrics: Optional[OrderedDict[str, MetricBase]] = None,
        validation_metrics: Optional[
            Union[
                OrderedDict[str, MetricBase],
                List[Tuple[str, OrderedDict[str, MetricBase]]],
            ]
        ] = None,
        test_metrics: Optional[
            Union[
                OrderedDict[str, MetricBase],
                List[Tuple[str, OrderedDict[str, MetricBase]]],
            ]
        ] = None,
        optimizers_and_lr_schs: Union[dict, Callable] = None,
        callbacks: Optional[Sequence[pl.Callback]] = None,
        best_epoch_source: Optional[Union[Dict, List[Dict]]] = None,
        save_hyperparameters_kwargs: Optional[dict] = None,
        save_model: bool = False,
        save_arguments: bool = False,
        tensorboard_sep: str = ".",
        log_unit: str = None,
        **kwargs: dict,
    ):
        """
        :param model_dir: location for checkpoints and logs
        :param model: Pytorch model to use
        :param losses: dict of FuseMedML style losses
               Will be used for both train and validation unless validation_losses is specified.
        :param validation_losses: Optional, typically used when there are multiple validation dataloaders - each with a different loss
                                List of tuples (must keep the same validation dataloaders order). Each tuple built from validation_dataloader name and the corresponding losses
        :param train_metrics: dict of FuseMedML style metrics - used for training set
        :param validation_metrics: ordereddict of FuseMedML style metrics - used for validation set (must be different instances of metrics (from train_metrics!)
                                   In case of multiple validation dataloaders,  validation_metrics should be list of tuples (that keeps the same dataloaders list order),
                                   Each tuple built from validation dataloader name and corresponding metrics dict.
        :param test_metrics: ordereddict of FuseMedML style metrics - used for test set (must be different instances of metrics (from train_metrics)
                                   In case of multiple test dataloaders, test_metrics should be list of tuples (that keeps the same dataloaders list order),
                                   Each tuple built from test dataloader name and corresponding metrics dict.
        :param optimizers_and_lr_schs: either a callable that follows pl.LightningModule.configure_optimizers prototype or just the output of such a function.
        :param callbacks: see pl.LightningModule.configure_callbacks return value for details
        :param best_epoch_source: Create list of pl.callbacks that saves checkpoints using (pl.callbacks.ModelCheckpoint) and print per epoch summary (fuse.dl.lightning.pl_epoch_summary.ModelEpochSummary).
                                  Either a dict with arguments to pass to ModelCheckpoint or list dicts for multiple ModelCheckpoint callbacks (to monitor and save checkpoints for more then one metric).
        :param save_hyperparameters_kwargs: specify pl.LightningModule.save_hyperparameters() arguments to save the hyper parameters. By default saved all except model_dir and the model (which stored separately)
                To load checkpoint (assuming save_model == True and save_arguments == True) do the following:
                "
                    model_dir = <path/to/the/original/model_dir>
                    checkpoint_path = os.path.join(model_dir, "last_epoch.ckpt")
                    nn_model = torch.load(os.path.join(model_dir, "model.pth"))
                    arguments = torch.load(os.path.join(model_dir, "arguments.pth"))
                    pl_model = LightningModuleDefault.load_from_checkpoint(checkpoint_path=checkpoint_path, model_dir=model_dir, model=nn_model, **arguments)
                "
                To load only the model (assuming save_model == True):
                "
                    model_dir = <path/to/the/original/model_dir>
                    checkpoint_path = os.path.join(model_dir, "last_epoch.ckpt")
                    nn_model = torch.load(os.path.join(model_dir, "model.pth"))
                    pl_model = LightningModuleDefault.load_from_checkpoint(checkpoint_path=checkpoint_path, model_dir=model_dir, model=nn_model)
                "

        :param save_model: save pickled format of the model
        :param save_arguments: save pickled format of main __init__ arguments (not including the model)
        :param tensorboard_sep: use "/" for cleaner tensorboard. "." is for backward compatibility.
        """
        super().__init__(**kwargs)

        if (save_arguments or save_model) and (model_dir is None):
            raise Exception(
                "Error: saving arguments or saving model requires a model_dir to be supplied as well."
            )

        # create model_dir
        if model_dir is not None:
            create_dir(model_dir)

        # save hyper parameters
        if save_hyperparameters_kwargs is not None:
            self.save_hyperparameters(**save_hyperparameters_kwargs)
        else:
            # do nothing - save the arguments once in model dir instead of on every checkpoint
            pass

        # save the model into model_dir - useful in case you want to load it without the original script
        if save_model:
            torch.save(model, os.path.join(model_dir, "model.pth"))

        # save the rest of the arguments - useful in case you want to load it without the original script
        if save_arguments:
            arguments = dict(
                losses=losses,
                train_metrics=train_metrics,
                validation_metrics=validation_metrics,
                test_metrics=test_metrics,
                optimizers_and_lr_schs=optimizers_and_lr_schs,
                callbacks=callbacks,
                best_epoch_source=best_epoch_source,
            )
            torch.save(arguments, os.path.join(model_dir, "arguments.pth"))

        # store arguments
        self._model_dir = model_dir
        self._model = model
        self._losses = losses if losses is not None else {}
        self._validation_losses = validation_losses

        self._train_metrics = train_metrics if train_metrics is not None else {}
        self._validation_metrics = (
            validation_metrics if validation_metrics is not None else {}
        )
        if test_metrics is None:
            self._test_metrics = self._validation_metrics
        else:
            self._test_metrics = test_metrics
        # convert all use-cases to the same format that supports multiple val dataloaders: List[Tuple[str, OrderedDict[str, MetricBase]]]
        if isinstance(self._validation_metrics, dict):
            self._validation_metrics = [(None, self._validation_metrics)]
        if isinstance(self._test_metrics, dict):
            self._test_metrics = [(None, self._test_metrics)]

        if log_unit not in [None, "optimizer_step", "epoch"]:
            raise Exception(f"Error: unexpected log_unit {log_unit}")

        self._log_unit = log_unit

        self._optimizers_and_lr_schs = optimizers_and_lr_schs
        self._callbacks = callbacks if callbacks is not None else []

        self._callbacks += model_default_callbacks(model_dir, best_epoch_source)

        # init state
        self._prediction_keys = None
        self._sep = tensorboard_sep

        self._training_step_outputs = []

        self._validation_step_outputs = {
            i: [] for i, _ in enumerate(self._validation_metrics)
        }
        self._test_step_outputs = {i: [] for i, _ in enumerate(self._test_metrics)}

    ## forward
    def forward(self, batch_dict: NDict) -> NDict:
        # workaround for fsdp
        if not isinstance(batch_dict, NDict):
            batch_dict = NDict(batch_dict)
        return self._model(batch_dict)

    ## Step
    def training_step(self, batch_dict: NDict, batch_idx: int) -> torch.Tensor:
        # add step number to batch_dict
        batch_dict["global_step"] = self.global_step
        # run forward function and store the outputs in batch_dict["model"]
        batch_dict = self.forward(batch_dict)
        # workaround for fsdp
        if not isinstance(batch_dict, NDict):
            batch_dict = NDict(batch_dict)
        # given the batch_dict and FuseMedML style losses - compute the losses, return the total loss and save losses values in batch_dict["losses"]
        total_loss = step_losses(self._losses, batch_dict)
        # given the batch_dict and FuseMedML style metrics - collect the required values to compute the metrics on epoch_end
        step_metrics(self._train_metrics, batch_dict)
        # aggregate losses
        self._training_step_outputs.append({"losses": batch_dict["losses"]})
        # return the total_loss
        return total_loss

    def validation_step(
        self, batch_dict: NDict, batch_idx: int, dataloader_idx: int = 0
    ) -> None:
        # add step number to batch_dict
        batch_dict["global_step"] = self.global_step
        # run forward function and store the outputs in batch_dict["model"]
        batch_dict = self.forward(batch_dict)
        # workaround for fsdp
        if not isinstance(batch_dict, NDict):
            batch_dict = NDict(batch_dict)
        if self._validation_losses is not None:
            losses = self._validation_losses[dataloader_idx][1]
        else:
            losses = self._losses
        # given the batch_dict and FuseMedML style losses - compute the losses, return the total loss (ignored) and save losses values in batch_dict["losses"]
        _ = step_losses(losses, batch_dict)
        # given the batch_dict and FuseMedML style metrics - collect the required values to compute the metrics on epoch_end
        step_metrics(self._validation_metrics[dataloader_idx][1], batch_dict)
        # aggregate losses
        if losses:  # if there are losses, collect the results
            self._validation_step_outputs[dataloader_idx].append(
                {"losses": batch_dict["losses"]}
            )

    def test_step(
        self, batch_dict: NDict, batch_idx: int, dataloader_idx: int = 0
    ) -> None:
        # add step number to batch_dict
        batch_dict["global_step"] = self.global_step
        # run forward function and store the outputs in batch_dict["model"]
        batch_dict = self.forward(batch_dict)
        # workaround for fsdp
        if not isinstance(batch_dict, NDict):
            batch_dict = NDict(batch_dict)
        # given the batch_dict and FuseMedML style losses - compute the losses, return the total loss (ignored) and save losses values in batch_dict["losses"]
        if self._validation_losses is not None:
            losses = self._validation_losses[dataloader_idx][1]
        else:
            losses = self._losses

        _ = step_losses(losses, batch_dict)
        # given the batch_dict and FuseMedML style metrics - collect the required values to compute the metrics on epoch_end
        step_metrics(self._test_metrics[dataloader_idx][1], batch_dict)
        # aggregate losses
        if losses:  # if there are losses, collect the results
            self._test_step_outputs[dataloader_idx].append(
                {"losses": batch_dict["losses"]}
            )

    def predict_step(self, batch_dict: NDict, batch_idx: int) -> dict:
        if self._prediction_keys is None:
            raise Exception(
                "Error: predict_step expects list of prediction keys to extract from batch_dict. Please specify it using set_predictions_keys() method "
            )
        # run forward function and store the outputs in batch_dict["model"]
        batch_dict = self.forward(batch_dict)
        # workaround for fsdp
        if not isinstance(batch_dict, NDict):
            batch_dict = NDict(batch_dict)
        # extract the required keys - defined in self.set_predictions_keys()
        return step_extract_predictions(self._prediction_keys, batch_dict)

    ## Epoch end
    def on_train_epoch_end(self) -> None:
        step_outputs = self._training_step_outputs
        # for the logs to be at each epoch, not each step
        if self._log_unit == "epoch":
            self.log("step", float(self.current_epoch), on_epoch=True, sync_dist=True)
        # calc average epoch loss and log it
        epoch_end_compute_and_log_losses(
            self, "train", [e["losses"] for e in step_outputs], sep=self._sep
        )
        # evaluate  and log it
        epoch_end_compute_and_log_metrics(
            self, "train", self._train_metrics, sep=self._sep
        )
        # reset state
        self._training_step_outputs.clear()

    def on_validation_epoch_end(self) -> None:
        step_outputs_lst = self._validation_step_outputs
        # for the logs to be at each epoch, not each step
        if self._log_unit == "epoch":
            self.log("step", float(self.current_epoch), on_epoch=True, sync_dist=True)
        for dataloader_idx, step_outputs in step_outputs_lst.items():
            if len(self._validation_metrics) == 1:
                prefix = "validation"
            else:
                prefix = f"validation.{self._validation_metrics[dataloader_idx][0]}"
            # calc average epoch loss and log it
            epoch_end_compute_and_log_losses(
                self, prefix, [e["losses"] for e in step_outputs], sep=self._sep
            )
            # evaluate  and log it
            epoch_end_compute_and_log_metrics(
                self, prefix, self._validation_metrics[dataloader_idx][1], sep=self._sep
            )
        # reset state
        self._validation_step_outputs = {
            i: [] for i, _ in enumerate(self._validation_metrics)
        }

    def on_test_epoch_end(self) -> None:
        step_outputs_lst = self._test_step_outputs
        # for the logs to be at each epoch, not each step
        if self._log_unit == "epoch":
            self.log("step", float(self.current_epoch), on_epoch=True, sync_dist=True)
        for dataloader_idx, step_outputs in step_outputs_lst.items():
            if len(self._test_metrics) == 1:
                prefix = "test"
            else:
                prefix = f"test.{self._test_metrics[dataloader_idx][0]}"
            # calc average epoch loss and log it
            epoch_end_compute_and_log_losses(
                self, prefix, [e["losses"] for e in step_outputs], sep=self._sep
            )
            # evaluate and log it
            epoch_end_compute_and_log_metrics(
                self, prefix, self._test_metrics[dataloader_idx][1], sep=self._sep
            )
        # reset state
        self._test_step_outputs = {i: [] for i, _ in enumerate(self._test_metrics)}

    # configuration
    def configure_callbacks(self) -> Sequence[pl.Callback]:
        """train loop callbacks"""
        return self._callbacks

    def configure_optimizers(self) -> torch.optim.Optimizer:
        """See pl.LightningModule.configure_optimizers return value for all options"""
        if isinstance(self._optimizers_and_lr_schs, Callable):
            return self._optimizers_and_lr_schs(self)
        return self._optimizers_and_lr_schs

    def set_predictions_keys(self, keys: List[str]) -> None:
        """Define which keys to extract from batch_dict on prediction mode"""
        self._prediction_keys = keys



import click
from typing import Union, List
from fuse.data.tokenizers.modular_tokenizer.modular_tokenizer import ModularTokenizer


@click.command()
@click.argument(
    "tokenizer-path",
    # "-p",
    default="../pretrained_tokenizers/bmfm_modular_tokenizer",
    # help="the directory containing the modular tokenizer",
)
@click.option(
    "--added-tokens",
    default=None,
    help="list of tokens to add",
)
@click.option(
    "--output-path",
    "-o",
    default=None,
    help="path to write tokenizer in",
)
# # this needs to be run on all the related modular tokenizers
def main(
    tokenizer_path: str, output_path: Union[str, None], added_tokens: List[str]
) -> None:
    print(f"adding special tokens to {tokenizer_path}")
    if output_path is None:
        output_path = tokenizer_path
    else:
        print(f"output into  {output_path}")

    tokenizer = ModularTokenizer.load(path=tokenizer_path)

    # Update tokenizer with special tokens:
    added_tokens = added_tokens.split(",")
    tokenizer.update_special_tokens(
        added_tokens=added_tokens,
        save_tokenizer_path=output_path,
    )

    print("Fin")


if __name__ == "__main__":
    main()



import random
import numpy as np
from typing import Optional, Sequence

from fuse.data.ops.op_base import OpBase

from fuse.utils import NDict


class OpAugOneHot(OpBase):
    """
    Apply an augmentation for an one-hot encoding vector.
    Op supports the following modes:

    mode == "default":
        replace the current one-hot encoding idx with a given one.

        example of use with OpRandApply as a meta-op:
                (
                    OpRandApply(OpAugOneHot(), probability=0.05),
                    dict(key="data.input.clinical.encoding.sex", idx=2, mode="default"),
                ),
            will change the one-hot index to 2 with probability of 0.05 (Here index 2 is N/A).

    mode == "ranking":
        replace the current one-hot encoding vector with a new one where the '1's idx will be +=1 the original one.
        meaning, (0,0,1,0) will turn into (0,1,0,0) or (0,0,0,1) (with prob 0.5).

        example of use with OpRandApply as a meta-op:
                (
                    OpRandApply(OpAugOneHot(), probability=0.05),
                    dict(key="data.input.clinical.encoding.age", mode="ranking", freeze_indices=[6]),
                ),
            will change the index of the age encoding += 1 with a probability of 0.05, but do nothing if the one-hot encoding has a '1' in the 6th index.
    """

    def __call__(
        self,
        sample_dict: NDict,
        key: str,
        idx: Optional[int] = None,
        freeze_indices: Sequence[int] = [],
        mode: str = "default",
    ) -> NDict:
        """
        :param key: key for a one-hot numpy array vector.
        :param idx: in "default" mode: idx to be change to 1
        :param freeze_indices: in "ranking" mode: sequence of indices such that if one-hot vector has a '1' in one of those indices, the augmentation won't be executed.
        :param mode: see class desc
        """

        supported_modes = ["default", "ranking"]
        if mode not in supported_modes:
            raise Exception(
                f"mode ({mode}) should be in supported modes ({supported_modes})."
            )

        if mode != "default" and idx is not None:
            raise Exception("specify idx only in default mode")

        if mode == "default" and idx is None:
            raise Exception("in 'default' mode, idx must be provided.")

        one_hot = sample_dict[key]
        res_one_hot = np.zeros_like(one_hot)

        if mode == "ranking":
            idx = np.argmax(one_hot)  # Get the current one-hot value

            if idx in freeze_indices:  # do not augment
                return sample_dict

            # idx +- 1 with probability of 0.5
            idx = (idx + 1) if random.random() < 0.5 else (idx - 1)

        # make sure idx in range
        idx = max(idx, 0)
        idx = min(idx, len(one_hot) - 1)

        # set one-hot
        res_one_hot[idx] = 1
        sample_dict[key] = res_one_hot

        return sample_dict



"""
(C) Copyright 2021 IBM Corp.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
   http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Created on June 30, 2021
"""
import pandas as pd
from typing import Any, Dict
from collections import OrderedDict

from fuse.eval.metrics.sequence_gen.metrics_seq_gen_common import (
    MetricPerplexity,
    MetricCountSeqAndTokens,
)

from fuse.eval.evaluator import EvaluatorDefault


from fuse.data import PipelineDefault, DatasetDefault, OpReadDataframe, CollateDefault
import torch
from torch.utils.data.dataloader import DataLoader


def example_seq_gen_0(seed: int = 1234) -> Dict[str, Any]:
    """
    Example/Test for perplexity metric
    """
    # entire dataframe at once case
    torch.manual_seed(seed=seed)
    pred = torch.randn(1000, 7, 100)
    pred = torch.nn.functional.softmax(pred, dim=-1).numpy()
    target = torch.randint(0, 100, (1000, 7)).numpy()

    data = {
        "pred": list(pred),
        "label": list(target),
        "id": list(range(1000)),
    }
    data = pd.DataFrame(data)

    metrics = OrderedDict(
        [
            (
                "perplexity",
                MetricPerplexity(preds="pred", target="label"),
            )
        ]
    )

    evaluator = EvaluatorDefault()
    results = evaluator.eval(ids=None, data=data, metrics=metrics)
    print(results)

    return results


def example_seq_gen_1(seed: int = 1234) -> Dict[str, Any]:
    """
    Example/Test for perplexity metric - batch mode
    """
    # entire dataframe at once case
    torch.manual_seed(seed=seed)
    pred = torch.randn(1000, 7, 100)
    pred = torch.nn.functional.softmax(pred, dim=-1).numpy()
    target = torch.randint(0, 100, (1000, 7)).numpy()

    data = {
        "pred": list(pred),
        "label": list(target),
        "id": list(range(1000)),
    }
    data = pd.DataFrame(data)

    # Working with pytorch dataloader mode
    dynamic_pipeline = PipelineDefault(
        "test",
        [
            (OpReadDataframe(data, key_column="id"), dict()),
        ],
    )
    ds = DatasetDefault(sample_ids=len(data), dynamic_pipeline=dynamic_pipeline)
    ds.create()
    dl = DataLoader(ds, collate_fn=CollateDefault(), batch_size=100)
    metrics = OrderedDict(
        [("perplexity", MetricPerplexity(preds="pred", target="label"))]
    )

    evaluator = EvaluatorDefault()
    results = evaluator.eval(ids=None, data=dl, metrics=metrics, batch_size=0)
    print(results)

    return results


def example_seq_gen_2() -> Dict[str, Any]:
    """
    Example/Test for perplexity metric - batch mode
    """

    encoder_input_tokens = torch.arange(5000).reshape(10, 500)
    data = {
        "encoder_input_tokens": list(encoder_input_tokens),
        "id": list(range(10)),
    }
    data = pd.DataFrame(data)

    # Working with pytorch dataloader mode
    dynamic_pipeline = PipelineDefault(
        "test",
        [
            (OpReadDataframe(data, key_column="id"), dict()),
        ],
    )
    ds = DatasetDefault(sample_ids=len(data), dynamic_pipeline=dynamic_pipeline)
    ds.create()
    dl = DataLoader(ds, collate_fn=CollateDefault())
    metrics = OrderedDict(
        [
            (
                "count",
                MetricCountSeqAndTokens(
                    encoder_input="encoder_input_tokens", ignore_index=4999
                ),
            )
        ]
    )

    evaluator = EvaluatorDefault()
    results = evaluator.eval(ids=None, data=dl, metrics=metrics, batch_size=0)
    print(results)

    return results



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join("..", "..")))


# -- Project information -----------------------------------------------------


# Take from setup.py file
def grep_value_from_setup(key: str) -> str:
    import re

    setup_file = os.path.normpath(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..", "setup.py")
    )
    with open(setup_file, "r") as setup_file:
        for line in setup_file.readlines():
            value = re.search(".*{key}=(.*),*".format(key=key), line)
            if value:
                value = value.group(1)
                value = re.sub(r"^\W+", "", value)  # lstrip non alphanumeric
                value = re.sub(r"\W+$", "", value)  # rstrip non alphanumeric
                return value


project = grep_value_from_setup("name")
copyright = "2023"
author = grep_value_from_setup("author")

# The full version, including alpha/beta/rc tags

release = "latest"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",  # link to other documentations
    # 'nbsphinx',  # work with jupyter notebook. requires pip install nbsphinx
    # 'sphinx.ext.mathjax',		# required for math in Jupyter Notebook
    "m2r2",  # work with markdown files. requires pip install m2r2
    "sphinx.ext.napoleon",  # parse Google/numpy docstring
    # 'sphinx.ext.autosummary',	# Automated index.rst toctreee
]

# autosummary_generate = True		# create api docs with autosummary


intersphinx_mapping = {
    "numpy": ("http://docs.scipy.org/doc/numpy/", None),
    "scipy": ("http://docs.scipy.org/doc/scipy/reference/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "python": ("https://docs.python.org/{}.{}".format(*sys.version_info), None),
}

autoclass_content = "both"

numpydoc_show_class_members = False  # Suppress sphinx warnings when building numpy-doc

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = {
    ".rst": "restructuredtext",
    ".txt": "restructuredtext",
    ".md": "markdown",
}

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "*estimation.*_.rst",  # estimation old files which kept locally
    "*tests.*",  # remove tests from documentation
]

# https://www.sphinx-doc.org/en/master/usage/configuration.html?highlight=master_doc#confval-master_doc
master_doc = "index"  # what file includes the root toctree. Needed explicitly for readthedocs.org


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
on_rtd = os.environ.get("READTHEDOCS") == "True"
if on_rtd:
    html_theme = "sphinx_rtd_theme"
else:
    html_theme = "classic"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# # Add content to the navigation side bar:
html_sidebars = {
    "**": ["localtoc.html", "sourcelink.html", "searchbox.html", "globaltoc.html"]
}


# -- Custom scripts -------------------------------------------------


# Add module's README for each module HTML page
def add_modules_readme() -> str:
    def get_rst_file_name_from_package_source(package_source_path: str) -> str:
        rst_source_path = os.path.relpath(package_source_path, os.getcwd())
        rst_file_name = rst_source_path.split(os.sep)
        rst_file_name = [
            directory for directory in rst_file_name if not directory.startswith(".")
        ]
        rst_file_name += ["rst"]
        rst_file_name = ".".join(rst_file_name)
        rst_file_name = os.path.join(source_html_dir, rst_file_name)
        return rst_file_name

    def get_edited_rast_file(
        rst_source_path: str, include_text: str, remove_original_header: bool = True
    ) -> str:
        with open(rst_source_path, "r") as fh:
            rst_source_lines = fh.read().splitlines()
        if not any([include_text in line for line in rst_source_lines]):
            # Add link to readme only if not already exists
            rst_source_lines.insert(3, include_text)
            if remove_original_header:
                rst_source_lines = rst_source_lines[2:]  # remove existing header

        return rst_source_lines

    INCLUDE_TEXT = ".. mdinclude:: "

    README_FILE_NAME = "README.md"

    # source_code_dir = os.path.join("..", "..", "fuse-med-ml")
    source_html_dir = os.getcwd() + "/docs/source/"  # sphinx's docs source directory

    for dir_name, subdir_list, file_names in os.walk(os.getcwd()):
        if README_FILE_NAME in file_names:  # Current dir has a readme file
            # Get README file path:
            # source_path = os.path.relpath(dir_name, os.path.join(source_code_dir, ".."))
            source_path = os.path.normpath(dir_name)

            readme_file_path = os.path.join(source_path, README_FILE_NAME)

            # Construct the corresponding module rst file:
            rst_source_file = get_rst_file_name_from_package_source(source_path)

            # Edit the rst file to include the path to the readme:
            path_to_be_added = os.path.relpath(readme_file_path, os.getcwd())

            try:
                include_text = INCLUDE_TEXT + "../../" + path_to_be_added + "\n"
                content = get_edited_rast_file(rst_source_file, include_text, True)

                with open(rst_source_file, "w") as f:
                    for line in content:
                        f.write("{}\n".format(line))

            except FileNotFoundError:
                print("Could not find file {}".format(rst_source_file))


add_modules_readme()



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import torch
import torch.nn as nn
import torch.nn.functional as F

from typing import Dict, Tuple, Sequence, Optional
from fuse.dl.models.heads.common import ClassifierMLP


class Head1D(nn.Module):
    def __init__(
        self,
        head_name: str = "head_0",
        mode: str = None,  # "classification" or "regression"
        conv_inputs: Sequence[Tuple[str, int]] = None,
        num_outputs: int = 2,
        append_features: Optional[Sequence[Tuple[str, int]]] = None,
        layers_description: Sequence[int] = (256,),
        append_layers_description: Sequence[int] = tuple(),
        append_dropout_rate: float = 0.0,
        dropout_rate: float = 0.1,
    ) -> None:
        """
        head 1d.

        Output of a forward pass for classification:
        'model.logits.head_name' and 'outputs.head_name', both in shape [batch_size, num_outputs]
        Output of a forward pass for regression:
        'model.output.head_name' in shape [batch_size, num_outputs]

        :param head_name:                   batch_dict key
        :param mode:                        "classification" or "regression"
        :param conv_inputs:                 List of feature map inputs - tuples of (batch_dict key, channel depth)
                                            If multiple inputs are used, they are concatenated on the channel axis
                for example:
                conv_inputs=(('model.backbone_features', 193),)
        :param num_outputs:                 Number of output classes (in case of classification) or just num outputs in case of regression
        :param append_features:          Additional vector (one dimensional) inputs, concatenated just before the classifier module
        :param layers_description:          Layers description for the classifier module - sequence of hidden layers sizes
        :param dropout_rate:                Dropout rate for classifier module layers
        """
        super().__init__()

        self.head_name = head_name
        self.mode = mode
        assert conv_inputs is not None, "conv_inputs must be provided"
        self.conv_inputs = conv_inputs
        self.append_features = append_features

        self.features_size = sum([conv_input[1] for conv_input in self.conv_inputs])

        if append_features is not None:
            if len(append_layers_description) == 0:
                self.features_size += sum(
                    [post_concat_input[1] for post_concat_input in append_features]
                )
                self.append_features_module = nn.Identity()
            else:
                self.features_size += append_layers_description[-1]
                self.append_features_module = ClassifierMLP(
                    in_ch=sum(
                        [post_concat_input[1] for post_concat_input in append_features]
                    ),
                    num_classes=None,
                    layers_description=append_layers_description,
                    dropout_rate=append_dropout_rate,
                )

        self.head_module = ClassifierMLP(
            in_ch=self.features_size,
            num_classes=num_outputs,
            layers_description=layers_description,
            dropout_rate=dropout_rate,
        )

    def forward(self, batch_dict: Dict) -> Dict:
        conv_input = torch.cat(
            [batch_dict[conv_input[0]] for conv_input in self.conv_inputs], dim=1
        )
        global_features = conv_input

        if self.append_features is not None:
            features = torch.cat(
                [
                    batch_dict[append_feature[0]]
                    for append_feature in self.append_features
                ]
            )
            features = self.append_features_module(features)
            features = features.reshape(features.shape + (1, 1, 1))
            if self.conv_inputs is not None:
                global_features = torch.cat((global_features, features), dim=1)
            else:
                global_features = features
        if self.mode == "regression":
            prediction = self.head_module(global_features).squeeze(dim=1)
            batch_dict["model.output." + self.head_name] = prediction
        else:
            logits = self.head_module(
                global_features
            )  # --> res.shape = [batch_size, 2, 1, 1]

            cls_preds = F.softmax(logits, dim=1)

            batch_dict["model.logits." + self.head_name] = logits
            batch_dict["model.output." + self.head_name] = cls_preds

        return batch_dict



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import shutil
import tempfile
import unittest
from fuse.utils.multiprocessing.run_multiprocessed import run_in_subprocess

from fuseimg.datasets.mnist import MNIST

from fuse.utils.rand.seed import Seed

from fuse_examples.imaging.classification.mnist.simple_mnist_starter import run_train


def run_mnist(root: str) -> None:
    Seed.set_seed(
        0, False
    )  # previous test (in the pipeline) changed the deterministic behavior to True
    try:
        run_train(num_epochs=1)
    except:
        raise Exception("Training MNIST Failed.")


class ClassificationMnistTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.root = tempfile.mkdtemp()

    def test_template(self) -> None:
        run_in_subprocess(run_mnist, self.root)

    def tearDown(self) -> None:
        # Delete temporary directories
        shutil.rmtree(self.root)


if __name__ == "__main__":
    unittest.main(verbosity=2)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021
"""

from copy import deepcopy
import os
from typing import Optional, Dict, Union
from fuse.utils.misc.misc import get_pretty_dataframe

import pytorch_lightning as pl
from pytorch_lightning.callbacks import Checkpoint

from pytorch_lightning.utilities.rank_zero import rank_zero_only
import torch
import pandas as pd


class ModelEpochSummary(Checkpoint):
    """
        Model Checkpointing Display
        ===================

        Automatically display (print to screen and log to a file) best vs current epoch metrics and losses.

        Example:
    Stats for epoch: 9 (best epoch is 7 for source validation.metrics.accuracy!)
    ------------------------------------------------------------------------------------------
    |                             | Best Epoch (7)              | Current Epoch (9)           |
    ------------------------------------------------------------------------------------------
    | train.losses.cls_loss       | 0.1546                      | 0.1146                      |
    ------------------------------------------------------------------------------------------
    | train.losses.total_loss     | 0.1546                      | 0.1146                      |
    ------------------------------------------------------------------------------------------
    | train.metrics.accuracy      | 0.9545                      | 0.9655                      |
    ------------------------------------------------------------------------------------------
    | validation.losses.cls_loss  | 0.0720                      | 0.0920                      |
    ------------------------------------------------------------------------------------------
    | validation.losses.total_loss| 0.0720                      | 0.0920                      |
    ------------------------------------------------------------------------------------------
    | validation.metrics.accuracy | 0.9776                      | 0.9756                      |
    ------------------------------------------------------------------------------------------

    """

    def __init__(
        self,
        dirpath: Optional[str] = None,
        filename: Optional[str] = None,
        monitor: Optional[str] = None,
        mode: str = "min",
    ):
        """
        :param dirpath: location of log file
        :param filename: specify a filename. If not set, it will use f"epoch_summary_{monitor}.txt"
        :param monitor: the metric name to track. if not set, it will just save the last model.
        :param mode: either consider the "min" value to be best or the "max" value to be the best
        """
        super().__init__()
        self._monitor = monitor
        self._mode = mode
        self._dirpath = dirpath
        if filename is None:
            self._filename = (
                f"epoch_summary_{self._monitor.replace('/', '.')}.txt"
                if self._monitor is not None
                else "epoch_summary.txt"
            )
        self._best_epoch_metrics = None
        self._best_epoch_index = None

    @rank_zero_only
    def print_epoch_summary_table(
        self, epoch_metrics: dict, epoch_source_index: int
    ) -> None:
        """
        Generate, print and log the epoch summary table.
        Decorator makes sure it runs only once in a DDP strategy.
        """

        def get_value_as_float_str(
            dict: Dict[str, Union[float, int]], key: str
        ) -> Union[None, str]:
            val_as_str = "N/A"
            try:
                value = dict[key]
                val_as_str = "%.4f" % float(value)
            except:
                pass
            return val_as_str

        stats_table = pd.DataFrame(
            columns=[
                "",
                f"Best Epoch ({self._best_epoch_index})",
                f"Current Epoch ({epoch_source_index})",
            ]
        )
        idx = 0

        eval_keys = sorted(epoch_metrics.keys())
        for evaluator_name in eval_keys:
            current_str = get_value_as_float_str(epoch_metrics, evaluator_name)
            best_so_far_str = get_value_as_float_str(
                self._best_epoch_metrics, evaluator_name
            )

            stats_table.loc[idx] = [f"{evaluator_name}", best_so_far_str, current_str]
            idx += 1

        if self._best_epoch_index == epoch_source_index:
            epoch_title = f"Stats for epoch: {epoch_source_index} (Currently the best epoch for source {self._monitor}!)"
            print(epoch_title)
        else:
            epoch_title = f"Stats for epoch: {epoch_source_index} (Best epoch is {self._best_epoch_index} for source {self._monitor})"
            print(epoch_title)
        stats_as_str = get_pretty_dataframe(stats_table)
        print(stats_as_str)
        print(f"Model: {self._dirpath}")

        try:
            op = (
                "w" if epoch_source_index == 0 else "a"
            )  # we want to have all stats in one file per epoch
            with open(os.path.join(self._dirpath, self._filename), op) as sfile:
                sfile.write(epoch_title)
                sfile.write(stats_as_str)
        except Exception as error:
            print(
                f"Cannot write epoch summary to file {self._dirpath}/{self._filename}"
            )
            print(error)

    def on_train_epoch_end(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> None:
        """Print summary at the end of the validation stage."""
        monitor_candidates = monitor_candidates = deepcopy(trainer.callback_metrics)
        current_epoch_metrics = monitor_candidates
        if self._monitor is not None:
            monitor_op = {"min": torch.lt, "max": torch.gt}[self._mode]
            if self._best_epoch_metrics is None or monitor_op(
                current_epoch_metrics[self._monitor],
                self._best_epoch_metrics[self._monitor],
            ):
                self._best_epoch_metrics = current_epoch_metrics
                self._best_epoch_index = trainer.current_epoch
        self.print_epoch_summary_table(current_epoch_metrics, trainer.current_epoch)



import unittest
import os
import tempfile


from fuse.data.datasets.dataset_default import DatasetDefault
from fuse.data.datasets.caching.samples_cacher import SamplesCacher

from fuseimg.datasets.isic import ISIC


class TestPipelineCaching(unittest.TestCase):
    def test_basic_1(self) -> None:
        """
        Test basic imaging ops
        """
        tmpdir = tempfile.mkdtemp()
        isic_dir = os.environ["ISIC19_DATA_PATH"]

        static_pipeline = ISIC.static_pipeline(isic_dir)

        cache_dirs = [
            os.path.join(tmpdir, "cache_a"),
            os.path.join(tmpdir, "cache_b"),
        ]

        cacher = SamplesCacher("fuseimg_ops_testing_cache", static_pipeline, cache_dirs)

        sample_ids = ["ISIC_0072637", "ISIC_0072638", "ISIC_0072639"]

        ds = DatasetDefault(
            sample_ids,
            static_pipeline,
            cacher=cacher,
        )

        ds.create()


if __name__ == "__main__":
    unittest.main()



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import torch.nn as nn
from torch import Tensor
from typing import Tuple, Optional, Callable, List, Sequence, Type


class Conv3DSimple(nn.Conv3d):
    def __init__(
        self,
        in_planes: int,
        out_planes: int,
        midplanes: Optional[int] = None,
        stride: int = 1,
        padding: int = 1,
    ) -> None:
        super().__init__(
            in_channels=in_planes,
            out_channels=out_planes,
            kernel_size=(3, 3, 3),
            stride=stride,
            padding=padding,
            bias=False,
        )

    @staticmethod
    def get_downsample_stride(stride: int) -> Tuple[int, int, int]:
        return stride, stride, stride


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(
        self,
        inplanes: int,
        planes: int,
        conv_builder: Callable[..., nn.Module],
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
    ) -> None:
        midplanes = (inplanes * planes * 3 * 3 * 3) // (inplanes * 3 * 3 + 3 * planes)

        super().__init__()
        self.conv1 = nn.Sequential(
            conv_builder(inplanes, planes, midplanes, stride),
            nn.BatchNorm3d(planes),
            nn.ReLU(inplace=True),
        )
        self.conv2 = nn.Sequential(
            conv_builder(planes, planes, midplanes), nn.BatchNorm3d(planes)
        )
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: Tensor) -> Tensor:
        residual = x

        out = self.conv1(x)
        out = self.conv2(out)
        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class BasicStem(nn.Sequential):
    """The default conv-batchnorm-relu stem"""

    def __init__(
        self,
        in_channels: int = 3,
        out_channels: int = 64,
        kernel_size: Tuple[int, int, int] = (3, 7, 7),
        stride: Tuple[int, int, int] = (1, 2, 2),
    ):
        padding = tuple([x // 2 for x in kernel_size])
        super().__init__(
            nn.Conv3d(
                in_channels,
                out_channels,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                bias=False,
            ),
            nn.BatchNorm3d(out_channels),
            nn.ReLU(inplace=True),
        )


class BackboneResnet3D(nn.Module):
    """
    A slightly more configureable ResNet3D implementation.
    Default values are identical to the pytorch implementation.
    """

    def __init__(
        self,
        *,
        in_channels: int = 3,
        pool: bool = False,
        layers: List[int] = [2, 2, 2, 2],
        first_channel_dim: int = 64,
        first_stride: int = 1,
        stem_kernel_size: Sequence[int] = (3, 7, 7),
        stem_stride: Sequence[int] = (1, 2, 2),
        pretrained: bool = False,
        name: str = "r3d_18",
    ) -> None:
        """
        :param in_channels: number of channels in the image
        :param pool: whether to use global average pooling to reduce the spacial dimensions after the convolutional layers

        :param layers: number of resnet blocks in every layer
        :param first_channel_dim: number of channels in the first layer. Every layer increases the channels by x2.
        :param first_stride: the stride in the first layer. using 2 will downsample the image by x2 in every spatial axis.

        :param stem_kernel_size: kernel size for the stem (first convolution)
        :param stem_stride: stride of the stem in every axis.

        :param pretrained: if True loads the pretrained video resnet model.
        :param name: name of the model to load, relevant only if pretrained=True
        """
        super().__init__()
        self.inplanes = first_channel_dim
        self.stem = BasicStem(
            3, first_channel_dim, kernel_size=stem_kernel_size, stride=stem_stride
        )

        self.layer1 = self._make_layer(
            BasicBlock, Conv3DSimple, first_channel_dim, layers[0], stride=first_stride
        )
        self.layer2 = self._make_layer(
            BasicBlock, Conv3DSimple, first_channel_dim * 2, layers[1], stride=2
        )
        self.layer3 = self._make_layer(
            BasicBlock, Conv3DSimple, first_channel_dim * 4, layers[2], stride=2
        )
        self.layer4 = self._make_layer(
            BasicBlock, Conv3DSimple, first_channel_dim * 8, layers[3], stride=2
        )
        self.out_dim = first_channel_dim * 8
        self.avgpool = nn.AdaptiveAvgPool3d((1, 1, 1))
        self._pool = pool
        self._pretrained = pretrained

        if pretrained:
            print("Warning: not supported by new torchvision version")
            from torch.hub import load_state_dict_from_url
            from torchvision.models.video.resnet import model_urls

            state_dict = load_state_dict_from_url(model_urls[name])
            del state_dict["fc.weight"]  # as a backbone the fc in not necessary
            del state_dict["fc.bias"]
            self.load_state_dict(state_dict)
        else:
            for m in self.modules():
                if isinstance(m, nn.Conv3d):
                    nn.init.kaiming_normal_(
                        m.weight, mode="fan_out", nonlinearity="relu"
                    )
                    if m.bias is not None:
                        nn.init.constant_(m.bias, 0)
                elif isinstance(m, nn.BatchNorm3d):
                    nn.init.constant_(m.weight, 1)
                    nn.init.constant_(m.bias, 0)
                elif isinstance(m, nn.Linear):
                    nn.init.normal_(m.weight, 0, 0.01)
                    nn.init.constant_(m.bias, 0)

        if in_channels == 1:
            self.stem[0].in_channels = 1
            self.stem[0].weight = nn.Parameter(
                self.stem[0].weight.sum(dim=1, keepdim=True)
            )
        elif in_channels != 3:
            self.stem = BasicStem(
                in_channels,
                first_channel_dim,
                kernel_size=stem_kernel_size,
                stride=stem_stride,
            )

    def _make_layer(
        self,
        block: Type[BasicBlock],
        conv_builder: Type[Conv3DSimple],
        planes: int,
        blocks: int,
        stride: int = 1,
    ) -> nn.Sequential:
        downsample = None

        if stride != 1 or self.inplanes != planes * block.expansion:
            ds_stride = conv_builder.get_downsample_stride(stride)
            downsample = nn.Sequential(
                nn.Conv3d(
                    self.inplanes,
                    planes * block.expansion,
                    kernel_size=1,
                    stride=ds_stride,
                    bias=False,
                ),
                nn.BatchNorm3d(planes * block.expansion),
            )
        layers = []
        layers.append(block(self.inplanes, planes, conv_builder, stride, downsample))

        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, conv_builder))

        return nn.Sequential(*layers)

    def features(self, x: Tensor) -> Tensor:
        x = self.stem(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        if self._pool:
            x = self.avgpool(x)
            x = x.flatten(1)
        return x

    def forward(self, x: Tensor) -> Tensor:
        """
        Forward pass. 3D global classification given a volume
        :param x: Input volume. shape: [batch_size, channels, z, y, x]
        :return: if pool is True:  feature vector per sample. shape: [batch_size, first_channel_dim * 8].
                 if pool is False: a 5D tensor of shape [batch_size, first_channel_dim * 8, z, y, x]
                                   where z,y,x are x8 or x16 smaller than the original dim (depends on the strides used)
        """
        return self.features(x)



from typing import Union, Sequence, Optional
from fuse.data import DatasetDefault
from fuse.data.utils.split import dataset_balanced_division_to_folds
from fuseimg.datasets.stoic21 import STOIC21


def train_val_test_splits(
    paths: Optional[dict] = None, params: Optional[dict] = None
) -> Sequence[Sequence]:
    # split to folds randomly - temp
    dataset_all = STOIC21.dataset(
        paths["data_dir"], paths["cache_dir"], reset_cache=False
    )
    folds = dataset_balanced_division_to_folds(
        dataset=dataset_all,
        output_split_filename=paths["data_split_filename"],
        keys_to_balance=["data.gt.probSevere"],
        nfolds=params["train"]["data.num_folds"],
    )

    test_sample_ids = []
    for fold in params["train"]["data.validation_folds"]:
        test_sample_ids += folds[fold]
    splits = []
    for fold in params["train"]["data.train_folds"]:
        train_sample_ids = sum(
            [folds[f] for f in params["train"]["data.train_folds"] if f != fold], []
        )
        val_sample_ids = folds[fold]
        splits.append([train_sample_ids, val_sample_ids, test_sample_ids])

    return splits


def create_dataset(
    train_val_sample_ids: Union[Sequence, None] = None,
    paths: Optional[dict] = None,
    params: Optional[dict] = None,
) -> Sequence[DatasetDefault]:
    if train_val_sample_ids is None:
        splits = train_val_test_splits(paths, params)
        train_val_sample_ids = sum(splits[0][0:2], [])
        test_sample_ids = splits[0][2]
        train_val_dataset = STOIC21.dataset(
            paths["data_dir"],
            paths["cache_dir"],
            sample_ids=train_val_sample_ids,
            train=True,
        )  # note all of this dataset will get a "training" pipeline (with augmentations etc')
        test_dataset = STOIC21.dataset(
            paths["data_dir"],
            paths["cache_dir"],
            sample_ids=test_sample_ids,
            train=False,
        )
        return train_val_dataset, test_dataset
    else:
        train_dataset = STOIC21.dataset(
            paths["data_dir"],
            paths["cache_dir"],
            sample_ids=train_val_sample_ids[0],
            train=True,
        )
        validation_dataset = STOIC21.dataset(
            paths["data_dir"],
            paths["cache_dir"],
            sample_ids=train_val_sample_ids[1],
            train=False,
        )

    return train_dataset, validation_dataset



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

===============================

The most minimal MNIST classifier implementation that demonstrate end to end training using Fuse that we could make.
There is a lot more that one could do on top of this, including inference, evaluation, model checkpointing, etc. But
the goal of this script is simply to train a classifier and show how to wrap a regular pytorch model so it can be
trained "fuse" style. Meaning it accesses data using keys to a batch_dict. It also demonstrates how easy it is to
load datasets from fuse.
"""
import copy
from typing import OrderedDict, Any, Tuple, Optional
from jsonargparse import CLI

import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data.dataloader import DataLoader
from torch.utils.data import DistributedSampler
import pytorch_lightning as pl

from fuse.eval.metrics.classification.metrics_thresholding_common import (
    MetricApplyThresholds,
)
from fuse.eval.metrics.classification.metrics_classification_common import (
    MetricAccuracy,
)

from fuse.data.utils.collates import CollateDefault
from fuse.data.utils.samplers import BatchSamplerDefault
from fuse.dl.losses.loss_default import LossDefault
from fuse.dl.models.model_wrapper import ModelWrapSeqToDict
from fuse.dl.lightning.pl_module import LightningModuleDefault

from fuseimg.datasets.mnist import MNIST
from fuse_examples.imaging.classification.mnist import lenet

import torch


def perform_softmax(logits: Any) -> Tuple[torch.Tensor, torch.Tensor]:
    cls_preds = F.softmax(logits, dim=1)
    return logits, cls_preds


## Model Definition #####################################################
class FuseLitLenet(LightningModuleDefault):
    def __init__(
        self, model_dir: Optional[str], save_model: bool, lr: float, weight_decay: float
    ):
        """
        Initialize the model. We inheret from LightningModuleDefault to get functions necessary for lightning trainer.
        We also wrap the torch model so that it can train using keys from the fuse NDict batch dict.

        :param model_dir: location for checkpoints and logs.
        :param lr: learning rate for optimizer
        :param weight_decay: weight decay for Adam optimizer. This is a form of regularization that penalizes for learning large weights.
        """

        # wrap basic torch model to automatically read inputs from batch_dict and save its outputs to batch_dict
        model = ModelWrapSeqToDict(
            model=lenet.LeNet(),
            model_inputs=["data.image"],
            post_forward_processing_function=perform_softmax,
            model_outputs=[
                "model.logits.classification",
                "model.output.classification",
            ],
        )

        # losses and metrics to track
        losses = {
            "cls_loss": LossDefault(
                pred="model.logits.classification",
                target="data.label",
                callable=F.cross_entropy,
                weight=1.0,
            ),
        }
        train_metrics = OrderedDict(
            [
                (
                    "operation_point",
                    MetricApplyThresholds(pred="model.output.classification"),
                ),  # will apply argmax
                (
                    "accuracy",
                    MetricAccuracy(
                        pred="results:metrics.operation_point.cls_pred",
                        target="data.label",
                    ),
                ),
            ]
        )
        validation_metrics = copy.deepcopy(train_metrics)  # same as train metrics

        # optimizer and learning rate scheduler
        optimizer = optim.Adam(model.parameters(), lr=lr, weight_decay=weight_decay)
        lr_scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer)
        lr_sch_config = dict(
            scheduler=lr_scheduler, monitor="validation.losses.total_loss"
        )
        optimizers_and_lr_schs = dict(optimizer=optimizer, lr_scheduler=lr_sch_config)

        # initialize LightningModuleDefault with our module, losses, etc so that we can use the functions of LightningModuleDefault
        super().__init__(
            model_dir=model_dir,
            save_model=save_model,
            model=model,
            losses=losses,
            train_metrics=train_metrics,
            validation_metrics=validation_metrics,
            optimizers_and_lr_schs=optimizers_and_lr_schs,
        )


## Datamodule ###########################################################
class MNISTDataModule(pl.LightningDataModule):
    def __init__(
        self,
        data_dir: str,
        batch_size: int,
        num_workers: int,
        distributed: bool,
    ):
        super().__init__()
        self.data_dir = data_dir
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.distributed = distributed

    def setup(self, stage: str) -> None:
        self.mnist_train = MNIST.dataset(self.data_dir, train=True)
        self.mnist_val = MNIST.dataset(self.data_dir, train=False)

    def train_dataloader(self) -> DataLoader:
        sampler = DistributedSampler(self.mnist_train) if self.distributed else None
        batch_sampler = BatchSamplerDefault(
            sampler=sampler,
            dataset=self.mnist_train,
            mode="approx",
            balanced_class_name="data.label",
            num_balanced_classes=10,
            batch_size=self.batch_size,
            balanced_class_weights=None,
        )
        train_loader = DataLoader(
            dataset=self.mnist_train,
            batch_sampler=batch_sampler,
            collate_fn=CollateDefault(),
            num_workers=self.num_workers,
        )
        return train_loader

    def val_dataloader(self) -> DataLoader:
        val_loader = DataLoader(
            self.mnist_val,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            collate_fn=CollateDefault(),
        )
        return val_loader


## Training #############################################################
def run_train(
    data_dir: str = "_examples/mnist",
    batch_size: int = 100,
    num_workers: int = 8,
    num_epochs: int = 3,
    num_devices: int = 1,
    num_nodes: int = 1,
    accelerator: str = "gpu",
    ckpt_path: Optional[str] = None,
    lr: float = 1e-4,
    weight_decay: float = 0.001,
) -> None:
    # initialize model
    model = FuseLitLenet(
        model_dir=None,
        save_model=False,
        lr=lr,
        weight_decay=weight_decay,
    )

    # initialize data module
    distributed = True if num_devices > 1 else None
    mnist_dm = MNISTDataModule(
        data_dir=data_dir,
        batch_size=batch_size,
        num_workers=num_workers,
        distributed=distributed,
    )

    # A workaround to support multiple GPUs with a custom batch_sampler for both Lightning versions
    #       see: https://lightning.ai/pages/releases/2.0.0/#sampler-replacement
    kwargs = {}
    if distributed:
        if pl.__version__[0] == "2":
            kwargs["use_distributed_sampler"] = False
        else:
            kwargs["replace_sampler_ddp"] = False

    # create pl trainer
    pl_trainer = pl.Trainer(
        max_epochs=num_epochs,
        accelerator=accelerator,
        devices=num_devices,
        num_nodes=num_nodes,
        **kwargs,
    )

    # train model
    pl_trainer.fit(model, mnist_dm)


if __name__ == "__main__":
    CLI(run_train)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import torch.nn as nn
from torchvision.models.video.mvit import MViT, MViT_V2_S_Weights
from typing import Dict, List
from torch import Tensor
import torch
from torchvision.models.video.mvit import MSBlockConfig


class BackboneMViT(nn.Module):
    def __init__(
        self,
        in_channels: int = 3,  # can be 1
        input_dim: List[int] = [16, 224, 224],
        pretrained: bool = False,
    ) -> None:
        """
        Multiscale Vision Transformers (MViT) is based on connecting the seminal idea of multiscale feature hierarchies with transformer models.
        This MViT model is based on the MViTv2: Improved Multiscale Vision Transformers for Classification and Detection (https://arxiv.org/abs/2112.01526)
        and Multiscale Vision Transformers papers (https://arxiv.org/abs/2104.11227).
        :param in_channels: number of channels in the image
        :param input_dim: shape of each image in the format [depth, height, width]

        :param pretrained: if True loads the pretrained video mvit_v2 model.
        """
        super(BackboneMViT, self).__init__()
        config: Dict[str, List] = {
            "num_heads": [1, 2, 2, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 8, 8],
            "input_channels": [
                96,
                96,
                192,
                192,
                384,
                384,
                384,
                384,
                384,
                384,
                384,
                384,
                384,
                384,
                384,
                768,
            ],
            "output_channels": [
                96,
                192,
                192,
                384,
                384,
                384,
                384,
                384,
                384,
                384,
                384,
                384,
                384,
                384,
                768,
                768,
            ],
            "kernel_q": [
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
            ],
            "kernel_kv": [
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
                [3, 3, 3],
            ],
            "stride_q": [
                [1, 1, 1],
                [1, 2, 2],
                [1, 1, 1],
                [1, 2, 2],
                [1, 1, 1],
                [1, 1, 1],
                [1, 1, 1],
                [1, 1, 1],
                [1, 1, 1],
                [1, 1, 1],
                [1, 1, 1],
                [1, 1, 1],
                [1, 1, 1],
                [1, 1, 1],
                [1, 2, 2],
                [1, 1, 1],
            ],
            "stride_kv": [
                [1, 8, 8],
                [1, 4, 4],
                [1, 4, 4],
                [1, 2, 2],
                [1, 2, 2],
                [1, 2, 2],
                [1, 2, 2],
                [1, 2, 2],
                [1, 2, 2],
                [1, 2, 2],
                [1, 2, 2],
                [1, 2, 2],
                [1, 2, 2],
                [1, 2, 2],
                [1, 1, 1],
                [1, 1, 1],
            ],
        }

        block_setting = []
        for i in range(len(config["num_heads"])):
            block_setting.append(
                MSBlockConfig(
                    num_heads=config["num_heads"][i],
                    input_channels=config["input_channels"][i],
                    output_channels=config["output_channels"][i],
                    kernel_q=config["kernel_q"][i],
                    kernel_kv=config["kernel_kv"][i],
                    stride_q=config["stride_q"][i],
                    stride_kv=config["stride_kv"][i],
                )
            )
        self.model = MViT(
            tuple(input_dim[1:]),
            input_dim[0],
            block_setting,
            residual_pool=True,
            residual_with_cls_embed=False,
            rel_pos_embed=True,
            proj_after_attn=True,
            stochastic_depth_prob=0.2,
        )
        if pretrained:
            if input_dim == [16, 224, 224]:
                weights = MViT_V2_S_Weights.DEFAULT
                self.model.load_state_dict(weights.get_state_dict(progress=True))
            else:
                print("pretrained compatible only for input_dim = [16,224,224]")

        if in_channels == 1:
            self.model.conv_proj.in_channels = 1
            self.model.conv_proj.weight = torch.nn.Parameter(
                self.model.conv_proj.weight.sum(dim=1, keepdim=True)
            )
        self.model.head = nn.Identity()

    def forward(self, x: Tensor) -> Tensor:
        return self.model(x)



import os
from zipfile import ZipFile
from fuse.utils.file_io.file_io import create_dir
import wget
from typing import Hashable, Optional, Sequence, List, Tuple
import torch
import numpy as np
import pytorch_lightning as pl
from fuse.data import DatasetDefault
from fuse.data.ops.ops_cast import OpToTensor
from fuse.data.utils.sample import get_sample_id
from fuse.data.pipelines.pipeline_default import PipelineDefault
from fuse.data.ops.op_base import OpBase
from fuse.data.datasets.caching.samples_cacher import SamplesCacher
from fuse.data.ops.ops_aug_common import OpSample, OpRandApply
from fuse.data.ops.ops_read import OpReadDataframe
from fuse.data.ops.ops_common import OpLambda, OpOverrideNaN, OpConcat
from fuseimg.data.ops.color import OpToRange
from fuse.data.ops.ops_aug_tabular import OpAugOneHot
from fuse.data.utils.split import dataset_balanced_division_to_folds
from fuse.utils import NDict

from fuseimg.data.ops.image_loader import OpLoadImage
from fuseimg.data.ops.aug.color import OpAugColor, OpAugGaussian
from fuseimg.data.ops.aug.geometry import OpResizeTo, OpAugAffine2D
from fuse.utils.rand.param_sampler import Uniform, RandInt, RandBool
from fuse.data.utils.samplers import BatchSamplerDefault
from torch.utils.data import DataLoader
from fuse.data.utils.collates import CollateDefault


class OpISICSampleIDDecode(OpBase):
    def __call__(self, sample_dict: NDict) -> NDict:
        """
        decodes sample id into image file name
        """
        sid = get_sample_id(sample_dict)

        img_filename_key = "data.input.img_path"
        sample_dict[img_filename_key] = sid + ".jpg"

        return sample_dict


def derive_label(sample_dict: NDict) -> NDict:
    """
    Takes the sample's ndict with the labels as key:value and assigns to sample_dict['data.label'] the index of the sample's class.
    Also delete all the labels' keys from sample_dict.

    for example:
        If the sample contains {'MEL': 0, 'NV': 1, 'BCC': 0, 'AK': 0, ... }
        will assign, sample_dict['data.label'] = 1 ('NV's index).
        Afterwards the sample_dict won't contain the class' names & values.
    """
    classes_names = ISIC.CLASS_NAMES

    label = 0
    for idx, cls_name in enumerate(classes_names):
        if int(sample_dict[f"data.cls_labels.{cls_name}"]) == 1:
            label = idx

    sample_dict["data.label"] = label
    return sample_dict


class ISIC:
    """
    ISIC 2019 challenge to classify dermoscopic images and clinical data among nine different diagnostic categories.
    """

    # bump whenever the static pipeline modified
    DATASET_VER = 0

    CLASS_NAMES = ["MEL", "NV", "BCC", "AK", "BKL", "DF", "VASC", "SCC"]
    SEX_INDEX = {"male": 0, "female": 1, "N/A": 2}
    ANATOM_SITE_INDEX = {
        "anterior torso": 0,
        "upper extremity": 1,
        "posterior torso": 2,
        "lower extremity": 3,
        "lateral torso": 4,
        "head/neck": 5,
        "palms/soles": 6,
        "oral/genital": 7,
        "N/A": 8,
    }

    @staticmethod
    def download(
        data_path: str, sample_ids_to_download: Optional[Sequence[str]] = None
    ) -> None:
        """
        Download images and metadata from ISIC challenge.
        Doesn't download again if data exists.
        :param data_path: location to store the data
        :param sample_ids_to_downlad: use if you want to download subset of the data. None will download and extract all the data.
        """
        create_dir(data_path)
        create_dir(os.path.join(data_path, "ISIC2019"))
        path = os.path.join(data_path, "ISIC2019/ISIC_2019_Training_Input")
        if not os.path.exists(path):
            print(
                "\nExtract ISIC-2019 training input ... (this may take a few minutes)"
            )

            url = "https://isic-challenge-data.s3.amazonaws.com/2019/ISIC_2019_Training_Input.zip"
            filename = os.path.join(data_path, "ISIC_2019_Training_Input.zip")
            print(f"Extract to {path}")
            if not os.path.exists(filename):
                wget.download(
                    url, os.path.join(data_path, "ISIC_2019_Training_Input.zip")
                )

            if sample_ids_to_download is not None:
                members = [
                    os.path.join("ISIC_2019_Training_Input", m + ".jpg")
                    for m in sample_ids_to_download
                ]
            else:
                members = None

            with ZipFile(filename, "r") as zipObj:
                # Extract all the contents of zip file in current directory
                zipObj.extractall(
                    path=os.path.join(data_path, "ISIC2019"), members=members
                )

            print("Extracting ISIC-2019 training input: done")

        path = os.path.join(data_path, "ISIC2019/ISIC_2019_Training_GroundTruth.csv")

        if not os.path.exists(path):
            print("\nExtract ISIC-2019 training gt ... (this may take a few minutes)")

            url = "https://isic-challenge-data.s3.amazonaws.com/2019/ISIC_2019_Training_GroundTruth.csv"
            wget.download(
                url,
                os.path.join(data_path, "ISIC2019/ISIC_2019_Training_GroundTruth.csv"),
            )

            print("Extracting ISIC-2019 training gt: done")

        path = os.path.join(data_path, "ISIC2019/ISIC_2019_Training_Metadata.csv")

        if not os.path.exists(path):
            print("\nExtract ISIC-2019 metadata ... (this may take a few minutes)")

            url = "https://isic-challenge-data.s3.amazonaws.com/2019/ISIC_2019_Training_Metadata.csv"
            wget.download(
                url, os.path.join(data_path, "ISIC2019/ISIC_2019_Training_Metadata.csv")
            )

            print("Extracting ISIC-2019 metadata: done")

    @staticmethod
    def sample_ids(data_path: str) -> List[str]:
        """
        Gets the samples ids in trainset.
        """
        images_path = os.path.join(data_path, "ISIC2019/ISIC_2019_Training_Input")

        samples = [
            f.split(".")[0]
            for f in os.listdir(images_path)
            if f.split(".")[-1] == "jpg"
        ]
        return samples

    @staticmethod
    def static_pipeline(data_path: str) -> PipelineDefault:
        rename_cls_labels = {c: f"data.cls_labels.{c}" for c in ISIC.CLASS_NAMES}
        rename_cls_labels[
            "image"
        ] = "data.cls_labels.sample_id"  # also extract image (sample_id)
        rename_metadata = {
            "age_approx": "data.input.clinical.age_approx",
            "anatom_site_general": "data.input.clinical.anatom_site_general",
            "sex": "data.input.clinical.sex",
            "image": "data.input.clinical.sample_id",
        }

        static_pipeline = PipelineDefault(
            "static",
            [
                # Decoding sample ID
                (OpISICSampleIDDecode(), dict()),
                # Load Image
                (
                    OpLoadImage(
                        os.path.join(data_path, "ISIC2019/ISIC_2019_Training_Input")
                    ),
                    dict(key_in="data.input.img_path", key_out="data.input.img"),
                ),
                # Normalize Images to range [0, 1]
                (
                    OpToRange(),
                    dict(key="data.input.img", from_range=(0, 255), to_range=(0, 1)),
                ),
                # Read labels into sample_dict. Each class will have a different entry.
                (
                    OpReadDataframe(
                        data_filename=os.path.join(
                            data_path, "ISIC2019/ISIC_2019_Training_GroundTruth.csv"
                        ),
                        key_column="data.cls_labels.sample_id",
                        columns_to_extract=list(rename_cls_labels.keys()),
                        rename_columns=rename_cls_labels,
                    ),
                    dict(),
                ),
                # Read metadata into sample_dict
                (
                    OpReadDataframe(
                        data_filename=os.path.join(
                            data_path, "ISIC2019/ISIC_2019_Training_Metadata.csv"
                        ),
                        key_column="data.input.clinical.sample_id",
                        columns_to_extract=list(rename_metadata.keys()),
                        rename_columns=rename_metadata,
                    ),
                    dict(),
                ),
                (
                    OpOverrideNaN(),
                    dict(
                        key="data.input.clinical.anatom_site_general",
                        value_to_fill="N/A",
                    ),
                ),
                (
                    OpOverrideNaN(),
                    dict(key="data.input.clinical.sex", value_to_fill="N/A"),
                ),
                (
                    OpOverrideNaN(),
                    dict(key="data.input.clinical.age_approx", value_to_fill=-1.0),
                ),
                # Squeeze labels into sample_dict['data.label']
                (OpLambda(func=derive_label), dict()),
                # Encode meta-data
                (
                    OpEncodeMetaData(),
                    dict(
                        key_site="data.input.clinical.anatom_site_general",
                        key_sex="data.input.clinical.sex",
                        key_age="data.input.clinical.age_approx",
                        out_prefix="data.input.clinical.encoding",
                    ),
                ),
            ],
        )
        return static_pipeline

    @staticmethod
    def dynamic_pipeline(
        train: bool = False, append: Optional[Sequence[Tuple[OpBase, dict]]] = None
    ) -> PipelineDefault:
        """
        Get suggested dynamic pipeline. including pre-processing that might be modified and augmentation operations.
        :param train: add augmentation if True
        :param append: pipeline steps to append at the end of the suggested pipeline
        """

        dynamic_pipeline = [
            # Resize images to 300x300x3
            (
                OpResizeTo(channels_first=True),
                dict(
                    key="data.input.img",
                    output_shape=(300, 300, 3),
                    mode="reflect",
                    anti_aliasing=True,
                ),
            ),
            # Convert to tensor for the augmentation process
            (OpToTensor(), dict(key="data.input.img", dtype=torch.float)),
        ]

        if train:
            dynamic_pipeline += [
                # Image Augmentation
                # Augmentation
                (
                    OpSample(OpAugAffine2D()),
                    dict(
                        key="data.input.img",
                        rotate=Uniform(-180.0, 180.0),
                        scale=Uniform(0.9, 1.1),
                        flip=(RandBool(0.3), RandBool(0.3)),
                        translate=(RandInt(-50, 50), RandInt(-50, 50)),
                    ),
                ),
                # Color augmentation
                (
                    OpSample(OpAugColor()),
                    dict(
                        key="data.input.img",
                        gamma=Uniform(0.9, 1.1),
                        contrast=Uniform(0.85, 1.15),
                        add=Uniform(-0.06, 0.06),
                        mul=Uniform(0.95, 1.05),
                    ),
                ),
                # Add Gaussian noise
                (OpAugGaussian(), dict(key="data.input.img", std=0.03)),
                # Meta-data Augmentation
                # Drop age with prob
                (
                    OpRandApply(OpAugOneHot(), probability=0.05),
                    dict(key="data.input.clinical.encoding.sex", idx=2, mode="default"),
                ),
                # switch age class with prob, except from unknowns
                (
                    OpRandApply(OpAugOneHot(), probability=0.05),
                    dict(
                        key="data.input.clinical.encoding.age",
                        mode="ranking",
                        freeze_indices=[6],
                    ),
                ),
            ]

        final = [
            # concat tabular data to one vector and cast it to tensor
            (
                OpConcat(),
                dict(
                    keys_in=[
                        f"data.input.clinical.encoding.{c}"
                        for c in ["site", "sex", "age"]
                    ],
                    key_out="data.input.clinical.all",
                ),
            ),
            (OpToTensor(), dict(key="data.input.clinical.all", dtype=torch.float)),
        ]

        if append is not None:
            dynamic_pipeline += append

        dynamic_pipeline += final

        return PipelineDefault("dynamic", dynamic_pipeline)

    @staticmethod
    def dataset(
        data_path: str,
        cache_path: str,
        train: bool = False,
        reset_cache: bool = False,
        num_workers: int = 10,
        append_dyn_pipeline: Optional[Sequence[Tuple[OpBase, dict]]] = None,
        samples_ids: Optional[Sequence[Hashable]] = None,
    ) -> DatasetDefault:
        """
        Get cached dataset
        :param train: if true, apply augmentations in the dynamic pipeline
        :param reset_cache: set to True to reset the cache
        :param num_workers: number of processes used for caching
        :param append_dyn_pipeline: pipeline steps to append at the end of the suggested dynamic pipeline
        :param samples_ids: dataset including the specified samples_ids or None for all the samples.
        """
        # Download data if doesn't exist
        ISIC.download(data_path=data_path, sample_ids_to_download=samples_ids)

        if samples_ids is None:
            samples_ids = ISIC.sample_ids(data_path)

        static_pipeline = ISIC.static_pipeline(data_path)
        dynamic_pipeline = ISIC.dynamic_pipeline(train, append=append_dyn_pipeline)

        cacher = SamplesCacher(
            f"isic_cache_ver{ISIC.DATASET_VER}",
            static_pipeline,
            [cache_path],
            restart_cache=reset_cache,
            workers=num_workers,
        )

        my_dataset = DatasetDefault(
            sample_ids=samples_ids,
            static_pipeline=static_pipeline,
            dynamic_pipeline=dynamic_pipeline,
            cacher=cacher,
        )

        my_dataset.create()
        return my_dataset


class ISICDataModule(pl.LightningDataModule):
    """
    Example of a custom Lightning datamodule that use FuseMedML tools.
    """

    def __init__(
        self,
        data_dir: str,
        cache_dir: str,
        num_workers: int,
        batch_size: int,
        train_folds: List[int],
        validation_folds: List[int],
        infer_folds: List[int],
        split_filename: str,
        sample_ids: Optional[Sequence[Hashable]] = None,
        reset_cache: bool = False,
        reset_split: bool = False,
        use_batch_sampler: bool = True,
        verify_folds_total_size: bool = True,
    ):
        """
        :param data_dir: path to data directory
        :param cache_dir: path to cache directory
        :param num_workers: number of process to use
        :param batch_size: model's batch size
        :param train_folds: which folds will be used for training
        :param validation_folds: which folds will be used for validation
        :param infer_folds: which folds will be used for inference (final evaluation)
        :param split_filename: path to file that will contain the data's split to folds
        :param sample_ids: subset of the sample ids to use, if None - use all.
        :param reset_cache: set True to reset the cache data
        :param reset_split: set True to reset the split file
        :param use_batch_sample: set True to use Fuse's custom balanced batch sampler (see BatchSamplerDefault class)
        :param verify_folds_total_size: (bool), when true - if a split file was found, will verify the the number of samples in the folds sum up to the amount of samples in the given dataset
        """
        super().__init__()
        self._data_dir = data_dir
        self._cache_dir = cache_dir
        self._num_workers = num_workers
        self._batch_size = batch_size
        self._use_batch_sampler = use_batch_sampler

        # divide into balanced train, validation and evaluation folds
        self._train_ids = []
        self._validation_ids = []
        self._infer_ids = []

        all_data = ISIC.dataset(
            self._data_dir,
            self._cache_dir,
            num_workers=self._num_workers,
            train=False,
            reset_cache=reset_cache,
            samples_ids=sample_ids,
        )

        folds = dataset_balanced_division_to_folds(
            dataset=all_data,
            keys_to_balance=["data.label"],
            nfolds=len(train_folds + validation_folds + infer_folds),
            output_split_filename=split_filename,
            reset_split=reset_split,
            verify_folds_total_size=verify_folds_total_size,
        )

        for fold in train_folds:
            self._train_ids += folds[fold]

        for fold in validation_folds:
            self._validation_ids += folds[fold]

        for fold in infer_folds:
            self._infer_ids += folds[fold]

    def setup(self, stage: str) -> None:
        """
        creates datasets by stage
        called on every process in DDP

        :param stage: trainer stage
        """
        # assign train/val datasets
        if stage == "fit":
            self._train_dataset = ISIC.dataset(
                self._data_dir,
                self._cache_dir,
                num_workers=self._num_workers,
                train=True,
                samples_ids=self._train_ids,
            )
            self._validation_dataset = ISIC.dataset(
                self._data_dir,
                self._cache_dir,
                num_workers=self._num_workers,
                train=False,
                samples_ids=self._validation_ids,
            )

        # assign prediction (infer) dataset
        if stage == "predict":
            self._predict_dataset = ISIC.dataset(
                self._data_dir,
                self._cache_dir,
                num_workers=self._num_workers,
                train=False,
                samples_ids=self._infer_ids,
            )

    def train_dataloader(self) -> DataLoader:
        """
        returns train dataloader with class args
        """
        if self._use_batch_sampler:
            # Create a batch sampler for the dataloader
            batch_sampler = BatchSamplerDefault(
                dataset=self._train_dataset,
                balanced_class_name="data.label",
                num_balanced_classes=8,
                batch_size=self._batch_size,
                workers=self._num_workers,
                verbose=True,
            )
            batch_size = 1  # should not provide batch_size for custom batch_sampler (1 is default)

        else:
            batch_sampler = None
            batch_size = self._batch_size

        # Create dataloader
        train_dl = DataLoader(
            dataset=self._train_dataset,
            batch_sampler=batch_sampler,
            batch_size=batch_size,
            collate_fn=CollateDefault(),
            num_workers=self._num_workers,
        )

        return train_dl

    def val_dataloader(self) -> DataLoader:
        """
        returns validation dataloader with class args
        """
        # Create dataloader
        validation_dl = DataLoader(
            dataset=self._validation_dataset,
            collate_fn=CollateDefault(),
            num_workers=self._num_workers,
            batch_size=self._batch_size,
        )

        return validation_dl

    def predict_dataloader(self) -> DataLoader:
        """
        returns validation dataloader with class args
        """
        # Create dataloader
        predict_dl = DataLoader(
            dataset=self._predict_dataset,
            collate_fn=CollateDefault(),
            num_workers=self._num_workers,
            batch_size=self._batch_size,
        )
        return predict_dl


class OpEncodeMetaData(OpBase):
    """
    Encode ISIC 2019 meta-data
    """

    def __init__(self, items_to_encode: List[str] = ["site", "sex", "age"]):
        """
        :param  items_to_encode: which items will be encoded. by default takes all the options.
        """
        super().__init__()

        self._items_to_encode = items_to_encode

    def __call__(
        self,
        sample_dict: NDict,
        key_site: str,
        key_sex: str,
        key_age: str,
        out_prefix: str,
    ) -> NDict:
        """
        :param key_site: sample_dict's key for patient's anatom site data
        :param key_sex: sample_dict's key for patient's sex data
        :param key_age: sample_dict's key for patient's age data
        :param out_prefix: the encoded data will be located in sample_dict[f"{out_prefix}.{data_type}"]
        """

        # Encode anatom site into a one-hot vector of length 9
        # 8 anatom sites and 1 for N/A
        if "site" in self._items_to_encode:
            site = sample_dict[key_site]
            site_one_hot = np.zeros(len(ISIC.ANATOM_SITE_INDEX))
            if site in ISIC.ANATOM_SITE_INDEX:
                site_one_hot[ISIC.ANATOM_SITE_INDEX[site]] = 1

            sample_dict[f"{out_prefix}.site"] = site_one_hot

        # Encode sex into a one-hot vector of length 3
        # male, female, N/A
        if "sex" in self._items_to_encode:
            sex = sample_dict[key_sex]
            sex_one_hot = np.zeros(len(ISIC.SEX_INDEX))
            if sex in ISIC.SEX_INDEX:
                sex_one_hot[ISIC.SEX_INDEX[sex]] = 1

            sample_dict[f"{out_prefix}.sex"] = sex_one_hot

        # Encode age into one-hot vector 'u' with length 7 such that:
        # for i in (0, ..., 5), u[i] == 1 iff age in range (20*i, 20(i+1)),
        # and u[6] == 1 iff age has missing value (0> or 120<)
        # for examples:
        #   age 50 -> [0,0,1,0,0,0,0]
        #   age 90 -> [0,0,0,0,1,0,0]
        #   missing age -> [0,0,0,0,0,0,1]
        if "age" in self._items_to_encode:
            age = int(sample_dict[key_age])
            age_one_hot = np.zeros(7)

            encode_idx = age // 20 if (age > 0 and age < 120) else 6
            age_one_hot[encode_idx] = 1

            sample_dict[f"{out_prefix}.age"] = age_one_hot

        return sample_dict



from typing import Sequence
import numpy as np
import torch
from torch import nn, Tensor
from fuse.dl.models.backbones.backbone_transformer import Transformer


class ProjectPatchesTokenizer(nn.Module):
    """
    Projects a 1D/2D/3D images to tokens using patches
    Assumes to have one of the forms:
    (1) batch_size, channels, height
    (2) batch_size, channels, height, width
    (3) batch_size, channels, height, width, depth

    The output shape is always:
    batch_size, num_tokens, token_dim
    """

    def __init__(
        self,
        *,
        image_shape: Sequence[int],
        patch_shape: Sequence[int],
        channels: int,
        token_dim: int,
        strict: bool = True,
    ):
        super().__init__()
        assert len(image_shape) == len(
            patch_shape
        ), "patch and image must have identical dimensions"
        np_image_shape = np.array(image_shape)
        np_patch_shape = np.array(patch_shape)

        # when shapes not divisible and strict=False, handling is similar to pytorch strided convolution
        # see here: https://pytorch.org/docs/stable/generated/torch.nn.Conv1d.html#torch.nn.Conv1d
        if strict:
            assert (
                np_image_shape % np_patch_shape == 0
            ).all(), "Image dimensions must be divisible by the patch size."
        self.num_tokens = int(np.prod(np_image_shape // np_patch_shape))
        patch_shape = tuple(patch_shape)
        self.image_dim = len(image_shape)
        if self.image_dim == 1:
            self.proj_layer = nn.Conv1d(
                in_channels=channels,
                out_channels=token_dim,
                kernel_size=patch_shape,
                stride=patch_shape,
            )
        elif self.image_dim == 2:
            self.proj_layer = nn.Conv2d(
                in_channels=channels,
                out_channels=token_dim,
                kernel_size=patch_shape,
                stride=patch_shape,
            )
        elif self.image_dim == 3:
            self.proj_layer = nn.Conv3d(
                in_channels=channels,
                out_channels=token_dim,
                kernel_size=patch_shape,
                stride=patch_shape,
            )
        else:
            raise NotImplementedError("only supports 1D/2D/3D images")

    def forward(self, x: Tensor) -> Tensor:
        assert (
            len(x.shape) == self.image_dim + 2
        ), "input should be [batch, channels] + image_shape"
        x = self.proj_layer(x)
        x = x.flatten(
            start_dim=2, end_dim=-1
        )  # x.shape == (batch_size, token_dim, num_tokens)
        x = x.transpose(1, 2)  # x.shape == (batch_size, num_tokens, token_dim)
        return x


class ViT(nn.Module):
    """
    Projects a 1D/2D/3D image into tokens, and then runs it through a transformer
    """

    def __init__(
        self, token_dim: int, projection_kwargs: dict, transformer_kwargs: dict
    ):
        """
        :param token_dim: the dimension of each token in the transformer
        :param projection_kwargs: positional arguments for the ProjectPatchesTokenizer class
        :param transformer_kwargs: positional arguments for the Transformer class
        """
        super().__init__()
        self.projection_layer = ProjectPatchesTokenizer(
            token_dim=token_dim, **projection_kwargs
        )
        num_tokens = self.projection_layer.num_tokens
        self.transformer = Transformer(
            num_tokens=num_tokens, token_dim=token_dim, **transformer_kwargs
        )

    def forward(self, x: Tensor, pool: str = "none") -> Tensor:
        """
        :param pool: returns all tokens (pool='none'), only cls token (pool='cls') or the average token (pool='mean')
        """
        assert pool in ["none", "cls", "mean"]
        x = self.projection_layer(x)
        x = self.transformer(x)
        if pool == "cls":
            x = x[:, 0]
        if pool == "mean":
            x = x.mean(dim=1)
        return x


def vit_tiny(
    image_shape: Sequence[int] = (224, 224),
    patch_shape: Sequence[int] = (16, 16),
    channels: int = 3,
) -> nn.Module:
    token_dim = 192
    projection_kwargs = dict(
        image_shape=image_shape, patch_shape=patch_shape, channels=channels
    )
    transformer_kwargs = dict(
        depth=12,
        heads=3,
        mlp_dim=token_dim * 4,
        dim_head=64,
        dropout=0.0,
        emb_dropout=0.0,
    )
    return ViT(
        token_dim=token_dim,
        projection_kwargs=projection_kwargs,
        transformer_kwargs=transformer_kwargs,
    )


def vit_small(
    image_shape: Sequence[int] = (224, 224),
    patch_shape: Sequence[int] = (16, 16),
    channels: int = 3,
) -> nn.Module:
    token_dim = 384
    projection_kwargs = dict(
        image_shape=image_shape, patch_shape=patch_shape, channels=channels
    )
    transformer_kwargs = dict(
        depth=12,
        heads=6,
        mlp_dim=token_dim * 4,
        dim_head=64,
        dropout=0.0,
        emb_dropout=0.0,
    )
    return ViT(
        token_dim=token_dim,
        projection_kwargs=projection_kwargs,
        transformer_kwargs=transformer_kwargs,
    )


def vit_base(
    image_shape: Sequence[int] = (224, 224),
    patch_shape: Sequence[int] = (16, 16),
    channels: int = 3,
) -> nn.Module:
    token_dim = 768
    projection_kwargs = dict(
        image_shape=image_shape, patch_shape=patch_shape, channels=channels
    )
    transformer_kwargs = dict(
        depth=12,
        heads=12,
        mlp_dim=token_dim * 4,
        dim_head=64,
        dropout=0.0,
        emb_dropout=0.0,
    )
    return ViT(
        token_dim=token_dim,
        projection_kwargs=projection_kwargs,
        transformer_kwargs=transformer_kwargs,
    )


def usage_example() -> Tensor:
    vit = vit_tiny()
    # an example input to the model
    x = torch.zeros([1, 3, 224, 224])
    print(f"image is projected into {vit.projection_layer.num_tokens} tokens")
    pred = vit(x, pool="cls")
    print(f"output shape is: {pred.shape}")
    return pred


if __name__ == "__main__":
    usage_example()



from os.path import join, basename, dirname
import re


def add_base_prefix(filepath: str, prefix: str) -> str:
    """
    Useful simple helper function for adding a prefix to a file. The prefix is added to the file "basename",
        and the full path is returned.
    Example usage:

        ans = add_base_prefix('/a/b/c/wow.tsv', 'extra_info@')
        print(ans)
        >'/a/b/c/extra_info@wow.tsv'


    """
    return join(dirname(filepath), prefix + basename(filepath))


def change_extension(filepath: str, new_extension: str) -> str:
    """
    modifies [filepath] extension to be [new_extension]

    for example _change_extension('/a/b/c.sdf/ew/some_file.zip', '7zip')
    will return '/a/b/c.sdf/ew/some_file.7zip'
    """

    _basename = basename(filepath)
    last_dot = _basename.rfind(".")
    if last_dot < 0:
        return filepath + "." + new_extension
    ans = join(dirname(filepath), _basename[:last_dot] + "." + new_extension)
    return ans


def get_extension(filepath: str) -> str:
    """
    Returns the extension.
    For example - get_extension('/a/b/c/d/banana.txt') will return '.txt'
    """
    _basename = basename(filepath)
    last_dot = _basename.rfind(".")
    if last_dot < 0:
        return ""
    ans = _basename[last_dot:]
    return ans


def remove_extension(filepath: str) -> str:
    """
    Returns the filename without the extension
    For example: remove_extension('/a/b/c/d/asdf.txt') with return '/a/b/c/d/asdf'
    """

    _dirname = dirname(filepath)
    _basename = basename(filepath)
    if "." not in _basename:
        return filepath

    last_dot = _basename.rfind(".")
    ans = join(_dirname, _basename[:last_dot])
    return ans


def get_valid_filename(s: str) -> str:
    """
    modifies an input string into a string that is valid as a filename in linux
    """
    s = str(s).strip().replace(" ", "_")
    return re.sub(r"[\\:\"*?<>|]", "@", s)



from fuse.data.ops.op_base import OpBase
from fuse.utils import NDict
import numpy as np
import os
from copy import deepcopy
import random
import pydicom
from scipy.ndimage import zoom
import torch.nn.functional as F
from typing import Tuple, Union, List
from volumentations import Compose


class OpLoadData(OpBase):
    """
    loads 3D image from a folder of dicom files.
    Folder should have only dicom files and they should be names in the order of the sequence.
    """

    def __init__(self, path_key: str = None):
        super().__init__()
        self.path_key = path_key

    def __call__(self, sample_dict: NDict) -> Union[None, dict, List[dict]]:
        """ """
        folder_path = sample_dict[self.path_key]
        dicom_files = [f for f in os.listdir(folder_path)]

        # Sort the DICOM files based on their file names (assuming they are numbered in order)
        dicom_files.sort()
        # Load the first DICOM file to get the image dimensions
        file_path = os.path.join(folder_path, dicom_files[0])
        dicom = pydicom.dcmread(file_path)
        rows, cols = dicom.pixel_array.shape

        # Create an empty 3D NumPy array to store the DICOM series
        num_slices = len(dicom_files)
        dicom_array = np.zeros((num_slices, rows, cols), dtype=dicom.pixel_array.dtype)

        # Load each DICOM file and store it in the 3D array
        for i, file_name in enumerate(dicom_files):
            file_path = os.path.join(folder_path, file_name)
            dicom = pydicom.dcmread(file_path)
            dicom_array[i] = dicom.pixel_array

        sample_dict["img"] = dicom_array

        return sample_dict


class OpNormalizeMRI(OpBase):
    def __call__(
        self,
        sample_dict: NDict,
        key: str,
        to_range: Tuple[float, float],
    ) -> Union[None, dict, List[dict]]:

        img = sample_dict[key]
        img = np.clip(
            img, *(np.percentile(img, [0, 95]))
        )  # truncate the intensities to the range of 0.5 to 99.5 percentiles

        from_range_start = img.min()
        from_range_end = img.max()
        to_range_start = to_range[0]
        to_range_end = to_range[1]

        # shift to start at 0
        img -= from_range_start
        if (from_range_end - from_range_start) == 0:
            print("MRI bad range")
            return None
        # scale to be in desired range
        img *= (to_range_end - to_range_start) / (from_range_end - from_range_start)
        # shift to start in desired start val
        img += to_range_start

        sample_dict[key] = img

        return sample_dict


class OpResize3D(OpBase):
    def __call__(
        self,
        sample_dict: NDict,
        key: str,
        shape: Tuple[int, int, int],
        segmentation: bool = False,
    ) -> Union[None, dict, List[dict]]:
        depth, height, width = sample_dict[key].shape
        depth_factor = shape[0] / depth
        height_factor = shape[1] / height
        width_factor = shape[2] / width
        if segmentation:
            sample_dict[key] = zoom(
                sample_dict[key],
                (depth_factor, height_factor, width_factor),
                order=0,
                mode="nearest",
            )
        else:
            sample_dict[key] = zoom(
                sample_dict[key], (depth_factor, height_factor, width_factor), order=1
            )
        return sample_dict


class OpDinoCrops(OpBase):
    """
    This function takes in a dictionary containing an image and a key, and returns a new dictionary with multiple crops of the original image.

    Parameters:
        sample_dict (NDict): A dictionary containing an image and a key.
        key (str): The key of the image in the dictionary.
        n_crops (int): The number of crops to create.

    Returns:
        NDict: A new dictionary with multiple crops of the original image.
    """

    def __call__(
        self, sample_dict: NDict, key: str, n_crops: int
    ) -> Union[None, dict, List[dict]]:
        """ """
        img = sample_dict[key]
        for i in range(n_crops):
            sample_dict[f"crop_{i}"] = deepcopy(img)

        del sample_dict[key]

        # sample_dict["data.input.img"] = np.moveaxis((np.repeat(img[...,np.newaxis],3,-1)), -1, 0)
        return sample_dict


class OpRandomCrop(OpBase):
    """
    Resizes the input image or volume to the specified size.

    Args:
        sample_dict (NDict): A dictionary containing the input image or volume.
        key (str): The key of the input image or volume in the sample_dict.
        scale (tuple, optional): A tuple specifying the minimum and maximum scaling factors. Defaults to (0.4, 1.0).
        on_depth (bool, optional): A boolean flag indicating whether to crop also the depth dimension. Defaults to True.
        res_shape (list, optional): A list specifying the target resolution. If provided, overrides the scale parameter. Defaults to None.

    Returns:
        NDict: A dictionary containing the resized image or volume.
    """

    def __call__(
        self,
        sample_dict: NDict,
        key: str,
        scale: Tuple[float, float] = (0.4, 1.0),
        on_depth: bool = True,
        res_shape: List = None,
    ) -> Union[None, dict, List[dict]]:
        """ """
        if isinstance(key, str):
            key = [key]
        shape = sample_dict[key[0]].shape
        if scale is not None:
            scale = random.uniform(scale[0], scale[1])
            scaled_shape = [round(scale * d) for d in shape]
        elif res_shape is not None:
            scaled_shape = res_shape
        starts = [random.randint(0, f - s) for (f, s) in zip(shape, scaled_shape)]
        for k in key:
            img = sample_dict[k]
            assert len(img.shape) == 3
            if on_depth:
                sample_dict[k] = img[
                    starts[0] : starts[0] + scaled_shape[0],
                    starts[1] : starts[1] + scaled_shape[1],
                    starts[2] : starts[2] + scaled_shape[2],
                ]
            else:
                sample_dict[k] = img[
                    :,
                    starts[1] : starts[1] + scaled_shape[1],
                    starts[2] : starts[2] + scaled_shape[2],
                ]
        return sample_dict


class OpRandomFlip(OpBase):
    def __call__(self, sample_dict: NDict, key: str) -> Union[None, dict, List[dict]]:
        """ """
        if isinstance(key, str):
            key = [key]

        flip_depth = np.random.choice([True, False])
        flip_spatial = np.random.choice([True, False])
        for k in key:
            if flip_depth:
                sample_dict[k] = np.flip(sample_dict[k], axis=0).copy()

            if flip_spatial:
                # Flip both height and width to maintain square shape
                sample_dict[k] = np.flip(sample_dict[k], axis=(1, 2)).copy()

        return sample_dict


class OpVolumentation(OpBase):
    def __init__(self, compose: Compose):
        super().__init__()
        self.compose = compose

    def __call__(self, sample_dict: NDict, key: str) -> Union[None, dict, List[dict]]:
        img = {"image": sample_dict[key]}
        sample_dict[key] = self.compose(**img)["image"]
        return sample_dict


class OpMask3D(OpBase):
    def __init__(
        self,
        mask_percentage: float = 0.3,
        cuboid_size: Tuple[int, int, int] = [2, 2, 2],
    ) -> Union[None, dict, List[dict]]:
        super().__init__()
        self.mask_percentage = mask_percentage
        self.cuboid_size = cuboid_size

    def __call__(self, sample_dict: NDict, key: str) -> NDict:
        img = sample_dict[key]

        depth, height, width = img.shape

        # Calculate the number of cuboids needed to mask the specified percentage
        total_voxels = depth * height * width
        cuboid_volume = self.cuboid_size[0] * self.cuboid_size[1] * self.cuboid_size[2]
        num_cuboids = int((self.mask_percentage / 100) * total_voxels / cuboid_volume)

        masked_image = img.copy()

        for _ in range(num_cuboids):
            # Generate random starting coordinates for the cuboid
            z = np.random.randint(0, depth - self.cuboid_size[0] + 1)
            y = np.random.randint(0, height - self.cuboid_size[1] + 1)
            x = np.random.randint(0, width - self.cuboid_size[2] + 1)

            # Mask the cuboid region
            masked_image[
                z : z + self.cuboid_size[0],
                y : y + self.cuboid_size[1],
                x : x + self.cuboid_size[2],
            ] = 0

        sample_dict[key] = masked_image

        return sample_dict


class OpSegToOneHot(OpBase):
    def __init__(self, n_classes: int):
        super().__init__()
        self.n_classes = n_classes

    def __call__(self, sample_dict: NDict, key: str) -> Union[None, dict, List[dict]]:
        seg_tensor = sample_dict[key]
        seg_tensor = seg_tensor.squeeze(0).long()
        one_hot = F.one_hot(seg_tensor, num_classes=self.n_classes)
        one_hot = one_hot.permute(3, 0, 1, 2)
        sample_dict[key] = one_hot
        return sample_dict



"""
(C) Copyright 2023 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Jan 11, 2023

"""

import torch
import unittest
import random
from fuse.dl.models.backbones.backbone_transformer import (
    CrossAttentionTransformerEncoder,
)


class TestCrossAttentionTransformerEncoder(unittest.TestCase):
    def test_all_contexts(self) -> None:
        """
        test cross attention transformer for each of the three context options: "seq_a", "seq_b" and "both"
        """

        # model parameters
        model_params = {
            "emb_dim": 128,
            "num_tokens_a": 10000,
            "num_tokens_b": 20000,
            "max_seq_len_a": 512,
            "max_seq_len_b": 1024,
            "output_dim": 256,
            "kwargs_wrapper_a": dict(emb_dropout=0.1),
            "kwargs_wrapper_b": dict(emb_dropout=0.1),
            "kwargs_encoder_a": dict(layer_dropout=0.1),
            "kwargs_cross_attn": dict(cross_attn_tokens_dropout=0.1),
        }

        # test for each context case
        for context in ["seq_a", "seq_b", "both"]:
            model_params["context"] = context
            self.validate_model_with_params(model_params)

    def validate_model_with_params(self, model_params: dict) -> None:
        """
        Basic validation for the CrossAttentionTransformerEncoder model

        :param model_params: A dictionary of the model's parameters to validate
        """

        # init model
        model = CrossAttentionTransformerEncoder(**model_params)

        # init random sequences that don't exceed max sequences length
        seq_a_len = random.randint(0, model_params["max_seq_len_a"])
        seq_b_len = random.randint(0, model_params["max_seq_len_b"])
        batch_size = random.randint(1, 10)
        s1 = torch.randint(0, model_params["num_tokens_a"], (batch_size, seq_a_len))
        s2 = torch.randint(0, model_params["num_tokens_b"], (batch_size, seq_b_len))

        # processing sample
        output = model(s1, s2)

        # validation
        assert output.shape[0] == batch_size
        if output[:, 0].shape[1] != model_params["output_dim"]:
            raise Exception(
                f"Expected output dimension to be {model_params['output_dim']}, but got: {output.shape[1]}. used model parameters: {model_params}."
            )


if __name__ == "__main__":
    unittest.main()



from typing import List, Optional, Union
from fuse.eval.metrics.libs.stat import Stat
from fuse.eval.metrics.metrics_common import MetricDefault
import numpy as np
from sklearn.metrics import (
    mean_absolute_error,
    mean_squared_error,
    r2_score,
    root_mean_squared_error,
)


class MetricPearsonCorrelation(MetricDefault):
    def __init__(
        self, pred: str, target: str, mask: Optional[str] = None, **kwargs: dict
    ) -> None:
        super().__init__(
            pred=pred,
            target=target,
            mask=mask,
            metric_func=Stat.pearson_correlation,
            **kwargs,
        )


class MetricSpearmanCorrelation(MetricDefault):
    def __init__(
        self, pred: str, target: str, mask: Optional[str] = None, **kwargs: dict
    ) -> None:
        super().__init__(
            pred=pred,
            target=target,
            mask=mask,
            metric_func=Stat.spearman_correlation,
            **kwargs,
        )


class MetricMAE(MetricDefault):
    def __init__(
        self,
        pred: str,
        target: str,
        **kwargs: dict,
    ) -> None:
        """
        See MetricDefault for the missing params
        :param pred: scalar predictions
        :param target: ground truth scalar labels
        :param threshold: threshold to apply to both pred and target
        :param balanced: optionally to use balanced accuracy (from sklearn) instead of regular accuracy.
        """
        super().__init__(
            pred=pred,
            target=target,
            metric_func=self.mae,
            **kwargs,
        )

    def mae(
        self,
        pred: Union[List, np.ndarray],
        target: Union[List, np.ndarray],
        **kwargs: dict,
    ) -> float:
        return mean_absolute_error(y_true=target, y_pred=pred)


class MetricMSE(MetricDefault):
    def __init__(
        self,
        pred: str,
        target: str,
        **kwargs: dict,
    ) -> None:
        """
        Our implementation of standard MSE, current version of scikit dones't support it as a metric.
        See MetricDefault for the missing params
        :param pred: scalar predictions
        :param target: ground truth scalar labels
        :param threshold: threshold to apply to both pred and target
        :param balanced: optionally to use balanced accuracy (from sklearn) instead of regular accuracy.
        """
        super().__init__(
            pred=pred,
            target=target,
            metric_func=self.mse,
            **kwargs,
        )

    def mse(
        self,
        pred: Union[List, np.ndarray],
        target: Union[List, np.ndarray],
        **kwargs: dict,
    ) -> float:
        return mean_squared_error(y_true=target, y_pred=pred)


class MetricRMSE(MetricDefault):
    def __init__(
        self,
        pred: str,
        target: str,
        **kwargs: dict,
    ) -> None:
        """
        See MetricDefault for the missing params
        :param pred: scalar predictions
        :param target: ground truth scalar labels
        """
        super().__init__(
            pred=pred,
            target=target,
            metric_func=self.rmse,
            **kwargs,
        )

    def rmse(
        self,
        pred: Union[List, np.ndarray],
        target: Union[List, np.ndarray],
        **kwargs: dict,
    ) -> float:
        pred = np.array(pred).flatten()
        target = np.array(target).flatten()
        assert len(pred) == len(
            target
        ), f"Expected pred and target to have the dimensions but found: {len(pred)} elements in pred and {len(target)} in target"

        return root_mean_squared_error(y_true=target, y_pred=pred)


class MetricR2(MetricDefault):
    def __init__(
        self,
        pred: str,
        target: str,
        **kwargs: dict,
    ) -> None:
        """
        See MetricDefault for the missing params
        :param pred: scalar predictions
        :param target: ground truth scalar labels
        """
        super().__init__(
            pred=pred,
            target=target,
            metric_func=self.r2,
            **kwargs,
        )

    def r2(
        self,
        pred: Union[List, np.ndarray],
        target: Union[List, np.ndarray],
        **kwargs: dict,
    ) -> float:

        pred = np.array(pred).flatten()
        target = np.array(target).flatten()

        assert len(pred) == len(
            target
        ), f"Expected pred and target to have the dimensions but found: {len(pred)} elements in pred and {len(target)} in target"

        return r2_score(target, pred)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""


from typing import Optional

from fuse.eval.metrics.libs.survival import MetricsSurvival
from fuse.eval.metrics.metrics_common import MetricDefault


class MetricCIndex(MetricDefault):
    """
    Compute c-index (concordance index) score using lifelines
    """

    def __init__(
        self,
        pred: str,
        event_times: str,
        event_observed: str,
        time_unit: Optional[int] = 1,
        time_followup: Optional[int] = None,
        **kwargs: dict,
    ):
        """
        See MetricDefault for the missing params
        :param pred: a length-n iterable prediction scores
        :param event_times:  a length-n iterable event times, or censor times if event is not observed.
        :param event_observed: a length-n iterable event observed flags, 1 if observed, 0 if not (i.e. censored).
        """
        super().__init__(
            pred=pred,
            target=None,
            event_times=event_times,
            event_observed=event_observed,
            time_unit=time_unit,
            time_followup=time_followup,
            metric_func=MetricsSurvival.c_index,
            **kwargs,
        )


class MetricExpectedCIndex(MetricDefault):
    """
    Compute expected C-index (concordance index) score when given a survival distribution
    """

    def __init__(
        self,
        pred: str,
        event_times: str,
        event_observed: str,
        **kwargs: dict,
    ):
        """
        See MetricDefault for the missing params
        :param pred: a length-n iterable prediction scores
        :param event_times:  a length-n iterable event times, or censor times if event is not observed.
        :param event_observed: a length-n iterable event observed flags, 1 if observed, 0 if not (i.e. censored).
        """
        super().__init__(
            pred=pred,
            target=None,
            event_times=event_times,
            event_observed=event_observed,
            metric_func=MetricsSurvival.expected_cindex,
            **kwargs,
        )



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
from typing import Any, Callable, Dict, List, Sequence, Optional, Tuple

import numpy as np
import torch
from torch.utils.data._utils.collate import default_collate
import torch.nn.functional as F

from fuse.utils import NDict
from fuse.utils.data.collate import CollateToBatchList
from fuse.data import get_sample_id_key


class CollateDefault(CollateToBatchList):
    """
    Default collate_fn method to be used when creating a DataLoader.
    Will collate each value with PyTorch default collate.
    Special collates per key can be specified in special_handlers_keys
    sample_id key will be collected to a list.
    Few options to special handlers implemented in this class as static methods
    """

    def __init__(
        self,
        skip_keys: Sequence[str] = tuple(),
        keep_keys: Sequence[str] = tuple(),
        raise_error_key_missing: bool = True,
        special_handlers_keys: Optional[Dict[str, Callable]] = None,
        post_collate_special_handlers_keys: Optional[List[Callable]] = None,
        add_to_batch_dict: Optional[Dict[str, Any]] = None,
    ):
        """
        :param skip_keys: do not collect the listed keys
        :param keep_keys: specifies a list of keys to collect. See raise_error_key_missing argument (dealing with missing keys).
        :param special_handlers_keys: per key specify a callable which gets as an input list of values and convert it to a batch.
                                      The rest of the keys will be converted to batch using PyTorch default collate_fn()
                                      Example of such Callable can be seen in the CollateDefault.pad_all_tensors_to_same_size.
        :param post_collate_special_handlers_keys: specify a callable which gets the batch_dict as an input and applies post processing to the batched tensors.
        :param raise_error_key_missing: if False, will not raise an error if there are keys that do not exist in some of the samples. Instead will set those values to None.
        :param add_to_batch_dict: optional, fixed items to add to batch_dict
        """
        super().__init__(skip_keys, raise_error_key_missing)
        self._special_handlers_keys = {}
        if special_handlers_keys is not None:
            self._special_handlers_keys.update(special_handlers_keys)
        self._special_handlers_keys[
            get_sample_id_key()
        ] = CollateDefault.just_collect_to_list
        self._keep_keys = keep_keys
        self._add_to_batch_dict = add_to_batch_dict

        self._post_collate_special_handlers_keys = post_collate_special_handlers_keys

    def __call__(self, samples: List[Dict]) -> Dict:
        """
        collate list of samples into batch_dict
        :param samples: list of samples
        :return: batch_dict
        """
        batch_dict = NDict()

        # collect all keys
        if self._keep_keys:
            keys = self._keep_keys
        else:
            keys = self._collect_all_keys(samples)

        # collect values
        for key in keys:
            # skip keys
            if key in self._skip_keys:
                continue

            try:
                # collect values into a list
                (
                    collected_values,
                    has_error,
                    has_missing_values,
                ) = self._collect_values_to_list(samples, key)

                # batch values
                self._batch_dispatch(
                    batch_dict,
                    samples,
                    key,
                    has_error or has_missing_values,
                    collected_values,
                )
            except:
                print(f"Error: Failed to collect key {key}")
                raise

        if self._add_to_batch_dict is not None:
            batch_dict.update(self._add_to_batch_dict)

        if self._post_collate_special_handlers_keys is not None:
            for callable in self._post_collate_special_handlers_keys:
                callable(batch_dict)

        return batch_dict

    def _batch_dispatch(
        self,
        batch_dict: dict,
        samples: List[dict],
        key: str,
        has_error: bool,
        collected_values: list,
    ) -> None:
        """
        dispatch a key into collate function and save it into batch_dict
        :param batch_dict: batch dictionary to update
        :param samples: list of samples
        :param key: key to collate
        :param has_error: True, if the key is missing in one of the samples
        :param collected values: the values collected from samples
        :return: nothing - the new batch will be added to batch_dict
        """
        if has_error:
            # do nothing when error occurs
            batch_dict[key] = collected_values
        elif key in self._special_handlers_keys:
            # use special handler if specified
            batch_dict[key] = self._special_handlers_keys[key](collected_values)
        elif isinstance(
            collected_values[0],
            (torch.Tensor, np.ndarray, float, int, str, bytes),
        ):
            # batch with default PyTorch implementation
            batch_dict[key] = default_collate(collected_values)
        else:
            batch_dict[key] = collected_values

    @staticmethod
    def just_collect_to_list(values: List[Any]) -> List[Any]:
        """
        special handler doing nothing - will just keep the collected list
        """
        return values

    @staticmethod
    def pad_all_tensors_to_same_size(
        values: List[torch.Tensor],
        pad_val: float = 0.0,
        min_size_per_dim: Optional[Tuple] = None,
    ) -> torch.Tensor:
        """
        pad tensors and create a batch - the shape will be the max size per dim
        values: list of tensor - all should have the same number of dimensions
        pad_val: constant value for padding
        min_size_per_dim: defines, per dimension, the minimal size in the post-collated tensor (excluding the batch dimension, which you shouldn't provide)
            this can be useful to prevent OOM due to memory fragmentation
            for example, you can use min_size_per_dim = (1000, -1) to make sure that the final collated tensor first dimension is at least 1000

        :return: torch.stack of padded tensors
        """

        # verify all are tensor and that they have the same dim size
        assert isinstance(
            values[0], torch.Tensor
        ), f"Expecting just tensors, got {type(values[0])}"
        num_dims = len(values[0].shape)

        for value in values:
            assert isinstance(
                value, torch.Tensor
            ), f"Expecting just tensors, got {type(value)}"
            assert (
                len(value.shape) == num_dims
            ), f"Expecting all tensors to have the same dim size, got {len(value.shape)} and {num_dims}"

        # get max per dim - this is the ovserved actual max from the batch
        max_per_dim = np.amax(np.stack([value.shape for value in values]), axis=0)

        if min_size_per_dim is not None:
            assert isinstance(min_size_per_dim, tuple)
            if len(min_size_per_dim) < len(max_per_dim):
                raise Exception(
                    f" length for min_size_per_dim={min_size_per_dim} expected to be >= max_per_dim={max_per_dim} but found length {len(min_size_per_dim)} and {len(max_per_dim)} respectively ! "
                )
            assert all(
                [(x > 0) or (x == -1) for x in min_size_per_dim]
            ), "allowed values for elements in min_size_per_dim are only positive integer or -1"
            max_per_dim = [
                max(actual, minimal_requested) if minimal_requested != -1 else actual
                for (actual, minimal_requested) in zip(max_per_dim, min_size_per_dim)
            ]

        # pad
        def _pad_size(value: torch.Tensor, dim: int) -> List[int]:
            assert max_per_dim[dim] >= value.shape[dim]
            return [0, max_per_dim[dim] - value.shape[dim]]

        padded_values = []

        for value in values:
            padding = []
            # F.pad padding description is expected to be provided in REVERSE order (see torch.nn.functional.pad doc)
            for dim in reversed(range(num_dims)):
                padding += _pad_size(value, dim)
            padded_value = F.pad(value, padding, mode="constant", value=pad_val)
            padded_values.append(padded_value)

        return default_collate(padded_values)

    @staticmethod
    def crop_padding(
        input_ids_list: List[torch.Tensor], pad_token_id: int
    ) -> torch.Tensor:
        """
        Crop padding of a batch of input_ids 1D tensors to the minimum length possible.

        Args:
            input_ids_list (list of torch.Tensor): List of input_ids tensors, where each tensor represents a sequence.
            pad_token_id (int): ID of the padding token used in input_ids tensors.

        Returns:
            torch.Tensor: Batched and cropped input_ids tensor with padding removed to the maximum length.

        Example:
            >>> input_ids_list = [
            ...     torch.tensor([101, 2054, 2003, 0, 0, 0, 0, 0, 0, 0]),
            ...     torch.tensor([101, 2023, 2003, 1037, 1999, 0, 0, 0, 0, 0]),
            ...     torch.tensor([101, 2002, 0, 0, 0, 0, 0, 0, 0, 0]),
            ... ]
            >>> pad_token_id = 0
            >>> cropped_batch = crop_padding_to_max_length(input_ids_list, pad_token_id)
            >>> print(cropped_batch)
            tensor([[ 101, 2054, 2003, 0,    0],
                    [ 101, 2023, 2003, 1037, 1999],
                    [ 101, 2002, 0,    0,    0]])

        Note:
            This function assumes that the input_ids tensors are already padded, and it crops the sequences
            to the minimum length by removing trailing padding tokens.
        """
        min_length = max(
            len(ids) - (ids == pad_token_id).sum().item() for ids in input_ids_list
        )
        cropped_sequences = [ids[:min_length] for ids in input_ids_list]
        batched_sequences = torch.stack(cropped_sequences, dim=0)
        return batched_sequences

    @staticmethod
    def crop_length_to_match_target_key(
        batch_dict: dict, target_key: str, keys_to_match: List[str]
    ) -> None:
        """
        Match the 1st dimension (typically the sequence length) of tensors in batch_dict, specified by keys, to a target tensor.

        Args:
            :param batch_dict: input dictionary with batched tensors with expected shape (B, L, *) where B is the batch dimension and L is a sequential dimension. L may vary across the elements in the batch_dict. * is any number of additional dimensions.
            :param target_key: key in batch_dict of the tensors with desired sequential dimension, L'.
            :param keys_to_match: list of keys in batch_dict with tensors who's 1st dimension should be converted to match the target_key tensor. The tensors to match should have 1st dimenions L >= L'.

        Returns:
            None, elements are modified in the batch_dict.
        """
        target_length = batch_dict[target_key].shape[1]
        for key in keys_to_match:
            batch_dict[key] = batch_dict[key][:, :target_length]
